#
# Collective Knowledge
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#

import unittest
import sys
import os

ck = None           # Will be updated by CK (initialized CK kernel)
test_util = None    # Will be updated by CK (initialized CK test utils)

# Contains new kernel tests. Add new tests here!


class TestKernel(unittest.TestCase):

    def test_out(self):
        with test_util.tmp_sys():
            ck.out('test')
            self.assertEqual('test', sys.stdout.getvalue().strip())

    def test_eout(self):
        with test_util.tmp_sys():
            ck.eout('test')
            self.assertEqual('test', sys.stderr.getvalue().strip())

    def test_err(self):
        with test_util.tmp_sys():
            ck.err({'return': 2, 'error': 'test.'})
            self.assertEqual('Error: test.\nExit code: 2',
                             sys.stdout.getvalue().strip())

    def test_jerr(self):
        with test_util.tmp_sys():
            with self.assertRaises(KeyboardInterrupt):
                ck.jerr({'return': 2, 'error': 'test.'})
            self.assertEqual('Error: test.', sys.stdout.getvalue().strip())

    def test_safe_float(self):
        import math

        self.assertEqual(ck.safe_float(1, 0), 1.0)
        self.assertEqual(ck.safe_float('a', 0), 0)
        self.assertEqual(ck.safe_float('-5.35', 0), -5.35)
        self.assertEqual(ck.safe_float('Infinity', 0), float('inf'))
        self.assertTrue(math.isnan(ck.safe_float('nan', 0)))

    def test_safe_int(self):
        self.assertEqual(ck.safe_int(1, 0), 1)
        self.assertEqual(ck.safe_int('a', 0), 0)
        self.assertEqual(ck.safe_int('-5', 0), -5)

    def test_safe_get_val_from_list(self):
        self.assertEqual(ck.safe_get_val_from_list([1, 2], 0, 0), 1)
        self.assertEqual(ck.safe_get_val_from_list([1, 2], 1, 0), 2)
        self.assertEqual(ck.safe_get_val_from_list([], 1, 0), 0)

    def test_system_with_timeout(self):
        r = ck.system_with_timeout({'cmd': 'exit 2'})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual(2, r['return_code'])

        r = ck.system_with_timeout({'cmd': 'sleep 5', 'timeout': 0.05})
        self.assertEqual(8, r['return'])

    def test_run_and_get_stdout(self):
        r = ck.run_and_get_stdout(
            {'cmd': ['sh', '-c', 'echo abcd && >&2 echo err && exit 2']})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual(2, r['return_code'])
        self.assertEqual('abcd\n', r['stdout'])
        self.assertEqual('err\n', r['stderr'])

        r = ck.run_and_get_stdout(
            {'cmd': 'sh -c "echo abcd && >&2 echo err && exit 2"'})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual(2, r['return_code'])
        self.assertEqual('abcd\n', r['stdout'])
        self.assertEqual('err\n', r['stderr'])

    def test_get_from_dicts(self):
        d1 = {'a': 1, 'b': 2}
        d2 = {'c': 3}
        r = ck.get_from_dicts(d1, 'a', '', d2)
        self.assertEqual(d1, {'b': 2})
        self.assertEqual(d2, {'a': 1, 'c': 3})
        self.assertEqual(1, r)

        ck.get_from_dicts(d1, 'a', 4, d2)
        self.assertEqual(d1, {'b': 2})
        self.assertEqual(d2, {'a': 1, 'c': 3})
        self.assertEqual(1, r)

    def test_convert_iso_time(self):
        r = ck.convert_iso_time({'iso_datetime': 'invalid'})
        self.assertEqual(1, r['return'])

        r = ck.convert_iso_time({'iso_datetime': '2016-09-07T13:47:34.5'})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual(2016, r['datetime_obj'].year)

    def test_convert_str_key_to_int(self):
        self.assertEqual(5, ck.convert_str_key_to_int('5'))
        self.assertEqual(0, ck.convert_str_key_to_int('a'))

    def test_inp(self):
        with test_util.tmp_sys('test input'):
            r = ck.inp({'text': 'test output'})
            self.assertEqual('test output', sys.stdout.getvalue().strip())
            self.assertEqual('test input', r['string'])
            self.assertEqual(0, r['return'], r.get('error', None))

    def test_select(self):
        d = {
            'key0': {'name': 'n0', 'sort': 1},
            'key1': {'name': 'n1', 'sort': 0}
        }
        with test_util.tmp_sys('1\n'):
            r = ck.select({'dict': d, 'title': 'Select:'})

            self.assertEqual(
                'Select:\n\n0) n1\n1) n0\n\nMake your selection (or press Enter for 0):', sys.stdout.getvalue().strip())
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual('key0', r['string'])

        with test_util.tmp_sys('\n'):
            r = ck.select({'dict': d, 'title': 'Select:',
                           'error_if_empty': 'yes'})
            self.assertEqual(1, r['return'])
            self.assertEqual('selection is empty', r['error'])

        with test_util.tmp_sys('\n'):
            r = ck.select({'dict': d, 'title': 'Select:'})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual('key1', r['string'])

        with test_util.tmp_sys('2\n'):
            r = ck.select({'dict': d, 'title': 'Select:'})
            self.assertEqual(1, r['return'])
            self.assertEqual('selection is not recognized', r['error'])

    def test_select_uoa(self):
        lst = [
            {'data_uid': 'uid1', 'data_uoa': 'b'},
            {'data_uid': 'uid2', 'data_uoa': 'a'}
        ]
        with test_util.tmp_sys('1\n'):
            r = ck.select_uoa({'choices': lst})

            self.assertEqual(
                '0) a (uid2)\n1) b (uid1)\n\nSelect UOA (or press Enter for 0):', sys.stdout.getvalue().strip())
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual('uid1', r['choice'])

        with test_util.tmp_sys('2\n'):
            r = ck.select_uoa({'choices': lst})
            self.assertEqual(1, r['return'])
            self.assertEqual('number is not recognized', r['error'])

    def test_convert_str_tags_to_list(self):
        self.assertEqual([1, 2], ck.convert_str_tags_to_list([1, 2]))
        self.assertEqual(['1', '2'], ck.convert_str_tags_to_list('  1 , 2  '))

    def test_check_writing(self):
        with test_util.tmp_cfg('forbid_global_delete'):
            r = ck.check_writing({'delete': 'yes'})
            self.assertEqual(1, r['return'])

        with test_util.tmp_cfg('forbid_global_writing'):
            r = ck.check_writing({})
            self.assertEqual(1, r['return'])

        with test_util.tmp_cfg('forbid_writing_modules'):
            r = ck.check_writing({'module_uoa': 'module'})
            self.assertEqual(1, r['return'])

        with test_util.tmp_cfg('forbid_writing_to_default_repo'):
            r = ck.check_writing({'repo_uoa': 'default'})
            self.assertEqual(1, r['return'])

        with test_util.tmp_cfg('forbid_writing_to_local_repo'):
            r = ck.check_writing({'repo_uoa': 'local'})
            self.assertEqual(1, r['return'])

        with test_util.tmp_cfg('allow_writing_only_to_allowed'):
            r = ck.check_writing({'repo_uoa': 'default'})
            self.assertEqual(1, r['return'])

        r = ck.check_writing({'repo_uoa': 'default', 'delete': 'yes', 'repo_dict': {
                             'forbid_deleting': 'yes'}})
        self.assertEqual(1, r['return'])

    def test_gen_tmp_file(self):
        with test_util.tmp_file(suffix='.txt') as fname:
            # check we can write to the file
            with open(fname, 'w') as f:
                f.write('test')

        fname = ck.gen_tmp_file({})['file_name']
        self.assertIn(os.sep, fname)

        fname = ck.gen_tmp_file({'remove_dir': 'yes'})['file_name']
        self.assertNotIn(os.sep, fname)

    def test_gen_uid(self):
        r = ck.gen_uid({})
        self.assertEqual(0, r['return'], r.get('error', None))

        uid = r['data_uid']
        self.assertEqual(16, len(uid))

        chars = '0123456789abcdef'
        for c in uid:
            self.assertIn(c, chars)

    def test_is_uid(self):
        self.assertTrue(ck.is_uid('5eac2a24f4a98e90'))

        self.assertFalse(ck.is_uid('a'))
        self.assertFalse(ck.is_uid('5evc2a24f4a98e90'))

    def test_is_uoa(self):
        self.assertTrue(ck.is_uoa('5evac2a24f4a98e90'))
        self.assertTrue(ck.is_uoa(''))

        self.assertFalse(ck.is_uoa('ab#12'))
        self.assertFalse(ck.is_uoa('ab^12'))
        self.assertFalse(ck.is_uoa('5evc2*a24f4a98e90'))
        self.assertFalse(ck.is_uoa('5evc2?a24f4a98e90'))

    def test_prepare_special_info_about_entry(self):
        r = ck.prepare_special_info_about_entry({})
        self.assertEqual(0, r['return'], r.get('error', None))
#        keys = sorted(['engine', 'version', 'author', 'author_email', 'author_webpage', 'license', 'copyright', 'iso_datetime'])
        # min set of keys to check, since other keys can differ depending on user configuration
        keys = ['engine', 'version', 'iso_datetime']
        xkeys = []

        xdict = r['dict'].keys()
        for k in keys:
            if k in xdict:
                xkeys.append(k)
        self.assertEqual(keys, xkeys)

    def test_convert_json_str_to_dict(self):
        r = ck.convert_json_str_to_dict({'str': "{'a': 1, 'b': [2, 3]}"})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual({'a': 1, 'b': [2, 3]}, r['dict'])

    def test_load_json_file(self):
        with test_util.tmp_file(suffix='.json', content='{"a": 1, "b": [2, 3]}') as fname:
            r = ck.load_json_file({'json_file': fname})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual({'a': 1, 'b': [2, 3]}, r['dict'])

    def test_load_yaml_file(self):
        with test_util.tmp_file(suffix='.yml', content='a: 1\nb:\n  - 2\n  - 3\n') as fname:
            r = ck.load_yaml_file({'yaml_file': fname})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual({'a': 1, 'b': [2, 3]}, r['dict'])

    def test_load_text_file(self):
        content = 'a\nb\nc'
        with test_util.tmp_file(suffix='.txt', content=content) as fname:
            r = ck.load_text_file({'text_file': fname, 'split_to_list': 'yes'})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual(str.encode(
                content.replace('\n', os.linesep)), r['bin'])
            self.assertEqual(content, r['string'])
            self.assertEqual(content.strip().split('\n'), r['lst'])

        content = 'a:1\nb:"z"\nc:2'
        with test_util.tmp_file(suffix='.txt', content=content) as fname:
            r = ck.load_text_file(
                {'text_file': fname, 'convert_to_dict': 'yes', 'remove_quotes': 'yes'})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual({'a': '1', 'b': 'z', 'c': '2'}, r['dict'])

    def test_substitute_str_in_file(self):
        content = 'a\nb\nc'
        with test_util.tmp_file(suffix='.txt', content=content) as fname:
            r = ck.substitute_str_in_file(
                {'filename': fname, 'string1': 'b', 'string2': 'd'})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual('a\nd\nc', ck.load_text_file(
                {'text_file': fname})['string'])

    def test_dumps_json(self):
        d = {'a': 1, 'b': [2, 3]}
        json_str = ck.dumps_json({'dict': d, 'skip_indent': 'yes'})['string']
        parsed_dict = ck.convert_json_str_to_dict(
            {'str': json_str, 'skip_quote_replacement': 'yes'})['dict']
        self.assertEqual(d, parsed_dict)

    def test_merge_dicts(self):
        d1 = {'a': 1, 'b': 2}
        d2 = {'c': 3, 'b': {}}
        ck.merge_dicts({'dict1': d1, 'dict2': d2})
        self.assertEqual({'a': 1, 'b': {}, 'c': 3}, d1)

    def test_save_yaml_to_file(self):
        d = {'a': 1, 'b': [2, 3]}
        with test_util.tmp_file(suffix='.yml') as fname:
            r = ck.save_yaml_to_file({'yaml_file': fname, 'dict': d})
            self.assertEqual(0, r['return'], r.get('error', None))

            r = ck.load_yaml_file({'yaml_file': fname})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual(d, r['dict'])

    def test_convert_file_to_upload_string(self):
        with test_util.tmp_file(suffix='.txt', content='http://ctuning.org/') as fname:
            r = ck.convert_file_to_upload_string({'filename': fname})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual('aHR0cDovL2N0dW5pbmcub3JnLw==',
                             r['file_content_base64'])

        with test_util.tmp_file() as fname:
            r = ck.convert_file_to_upload_string({'filename': fname})
            self.assertEqual(1, r['return'])

    def test_convert_upload_string_to_file(self):
        with test_util.tmp_file(suffix='.txt') as fname:
            r = ck.convert_upload_string_to_file(
                {'file_content_base64': 'aHR0cDovL2N0dW5pbmcub3JnLw==', 'filename': fname})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual(fname, r['filename'])

            r = ck.load_text_file({'text_file': fname})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual('http://ctuning.org/', r['string'])

        r = ck.convert_upload_string_to_file(
            {'file_content_base64': 'aHR0cDovL2N0dW5pbmcub3JnLw=='})
        fname = r['filename']
        try:
            self.assertEqual(0, r['return'], r.get('error', None))

            r = ck.load_text_file({'text_file': fname})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual('http://ctuning.org/', r['string'])
        finally:
            os.remove(fname)

        with test_util.tmp_file(suffix='.txt', content='test') as fname:
            r = ck.convert_upload_string_to_file(
                {'file_content_base64': 'aHR0cDovL2N0dW5pbmcub3JnLw==', 'filename': fname})
            self.assertEqual(1, r['return'])

    def test_convert_ck_list_to_dict(self):
        r = ck.convert_ck_list_to_dict(['a', '--', 1, '2'])
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual({'action': 'a', 'cids': [],
                          'unparsed': [1, '2']}, r['ck_dict'])

        r = ck.convert_ck_list_to_dict(['a', '-k=1'])
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual({'action': 'a', 'cids': [], 'k': '1'}, r['ck_dict'])

        r = ck.convert_ck_list_to_dict(['a', '@@@2'])
        self.assertEqual(1, r['return'])

        r = ck.convert_ck_list_to_dict(['a', '@@@{\'k\': 1}'])
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual({'action': 'a', 'cids': [], 'k': 1}, r['ck_dict'])

        with test_util.tmp_sys('{"k": 1}\n\n'):
            r = ck.convert_ck_list_to_dict(['a', '@@'])
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual({'action': 'a', 'cids': [], 'k': 1}, r['ck_dict'])

        with test_util.tmp_sys('{"k": 1}\n\n'):
            r = ck.convert_ck_list_to_dict(['a', '@@q'])
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual({'action': 'a', 'cids': [],
                              'q': {'k': 1}}, r['ck_dict'])

        r = ck.convert_ck_list_to_dict(['a', '@2'])
        self.assertEqual(1, r['return'])

        with test_util.tmp_file(suffix='.tmp', content='{"k": 1}') as fname:
            r = ck.convert_ck_list_to_dict(['a', '@' + fname])
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual({'action': 'a', 'cids': [], 'k': 1}, r['ck_dict'])
            self.assertFalse(os.path.isfile(fname))

    def test_input_json(self):
        content = '{"a": 1, "b": [2, 3]}\n\n'
        with test_util.tmp_sys(content):
            r = ck.input_json({'text': 'input json'})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual('input json', sys.stdout.getvalue().strip())
            self.assertEqual(content.strip(), r['string'])
            self.assertEqual({'a': 1, 'b': [2, 3]}, r['dict'])

    def test_list_all_files(self):
        with test_util.tmp_dir() as fname:
            os.makedirs(fname + '/a')
            os.makedirs(fname + '/b')

            with open(fname + '/a/test.txt', 'a') as f:
                f.write('abc')

            with open(fname + '/b/test.log', 'a') as f:
                f.write('12')

            s = os.sep

            r = ck.list_all_files({'path': fname, 'limit': 10})
            self.assertEqual(0, r['return'], r.get('error', None))
            t = r['list']
            self.assertEqual({'size': 3}, t['a' + s + 'test.txt'])
            self.assertEqual({'size': 2}, t['b' + s + 'test.log'])

            r = ck.list_all_files({'path': fname, 'pattern': '*.txt'})
            self.assertEqual(0, r['return'], r.get('error', None))
            t = r['list']
            self.assertEqual({'size': 3}, t['a' + s + 'test.txt'])
            self.assertEqual(None, t.get('b' + s + 'test.log'))

            r = ck.list_all_files({'path': fname, 'file_name': 'test.txt'})
            self.assertEqual(0, r['return'], r.get('error', None))
            t = r['list']
            self.assertEqual({'size': 3}, t['a' + s + 'test.txt'])

        with test_util.tmp_dir() as fname:
            with open(fname + '/test.txt', 'a') as f:
                f.write('abc')
            with open(fname + '/test.log', 'a') as f:
                f.write('12')
            with open(fname + '/test.abc', 'a') as f:
                f.write('12')

            r = ck.list_all_files({'path': fname, 'limit': 1})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual(1, len(r['list']))

            os.makedirs(fname + '/a')
            os.makedirs(fname + '/b')

            with open(fname + '/a/1.txt', 'a') as f:
                f.write('abc')

            with open(fname + '/b/1.log', 'a') as f:
                f.write('12')

            r = ck.list_all_files({'path': fname, 'limit': 1})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual(1, len(r['list']))

    def test_save_repo_cache(self):
        r = ck.save_repo_cache({})
        self.assertEqual(0, r['return'], r.get('error', None))

        r = ck.load_json_file({'json_file': ck.work['dir_cache_repo_uoa']})
        self.assertEqual(ck.cache_repo_uoa, r['dict'])

        r = ck.load_json_file({'json_file': ck.work['dir_cache_repo_info']})
        self.assertEqual(ck.cache_repo_info, r['dict'])

    def test_load_repo_info_from_cache(self):
        r = ck.load_repo_info_from_cache({'repo_uoa': 'default'})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('default', r['data_alias'])
        self.assertEqual('default', r['data_uoa'])

        missing_repo = ck.gen_uid({})['data_uid']
        r = ck.load_repo_info_from_cache({'repo_uoa': missing_repo})
        self.assertEqual(1, r['return'])

        r = ck.load_repo_info_from_cache({'repo_uoa': missing_repo + '='})
        self.assertEqual(1, r['return'])

    def test_find_repo_by_path(self):
        r = ck.find_path_to_repo({'repo_uoa': 'default'})
        r = ck.find_repo_by_path({'path': r['path']})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('default', r['repo_uoa'])
        self.assertEqual('default', r['repo_alias'])

        r = ck.find_path_to_repo({'repo_uoa': 'local'})
        r = ck.find_repo_by_path({'path': r['path']})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('local', r['repo_uoa'])
        self.assertEqual('local', r['repo_alias'])

        with test_util.tmp_file() as missing_repo_path:
            r = ck.find_repo_by_path({'path': missing_repo_path})
            self.assertEqual(16, r['return'])

        with test_util.tmp_repo() as r:
            path = r['dict']['path']
            name = r['data_name']
            r = ck.find_repo_by_path({'path': path})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual(name, r['repo_uoa'])
            self.assertEqual(name, r['repo_alias'])

    def test_find_path_to_repo(self):
        with test_util.tmp_repo() as r:
            path = r['dict']['path']
            name = r['data_name']
            r = ck.find_path_to_repo({'repo_uoa': name})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertEqual(path, r['path'])

        missing_repo = ck.gen_uid({})['data_uid']
        r = ck.find_path_to_repo({'repo_uoa': missing_repo + '='})
        self.assertEqual(1, r['return'])

    def test_find_path_to_data(self):
        missing_data = ck.gen_uid({})['data_uid']
        r = ck.find_path_to_data(
            {'module_uoa': 'default', 'data_uoa': missing_data})
        self.assertEqual(16, r['return'])

    def test_load_module_from_path(self):
        repo_path = ck.find_path_to_repo({'repo_uoa': 'local'})['path']
        missing_module = ck.gen_uid({})['data_uid']
        r = ck.load_module_from_path(
            {'path': repo_path, 'module_code_name': missing_module})
        self.assertEqual(1, r['return'])

        with test_util.tmp_dir() as dname:
            module_name = 'test'
            module_file = os.path.join(dname, module_name + '.py')
            with open(module_file, 'w') as f:
                f.write('pass\n')
            r = ck.load_module_from_path(
                {'path': dname, 'module_code_name': 'test', 'cfg': {'min_kernel_dep': '10'}})
            self.assertEqual(1, r['return'])

    def test_perform_action(self):
        r = ck.perform_action({})
        self.assertEqual(0, r['return'], r.get('error', None))

        with test_util.tmp_dir(cwd=True):
            r = ck.perform_action({'cid': '#'})
            self.assertEqual(16, r['return'])
            self.assertEqual(
                'repository is not detected in the current path', r['error'])

        with test_util.tmp_dir(cwd=True):
            r = ck.perform_action({'cids': ['#']})
            self.assertEqual(16, r['return'])
            self.assertEqual(
                'repository is not detected in the current path', r['error'])

        with test_util.tmp_sys():
            r = ck.perform_action({'cid': 'default:test:', 'cids': [
                                  'default:test:'], 'action': 'cmd'})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertTrue(sys.stdout.getvalue(
            ).strip().startswith('Command line:'))

    def test_get_api(self):
        r = ck.get_api({'func': 'list_actions'})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual(
            'List actions in the given CK module\n \nTARGET: should use via ck.kernel.access', r['title'].strip())
        self.assertTrue(
            r['api'].strip().lower().startswith('target audience:'))

        missing_file = ck.gen_uid({})['data_uid'] + '.py'
        with test_util.tmp_cfg('file_kernel_py', missing_file):
            r = ck.get_api({})
            self.assertEqual(1, r['return'])
            self.assertTrue(r['error'].strip().startswith(
                'kernel not found in'))

        r = ck.get_api({'module_uoa': 'test'})
        self.assertEqual(1, r['return'])
        self.assertEqual('function not found', r['error'])

#        with test_util.tmp_sys():
#            r = ck.get_api({'module_uoa': 'test', 'func': 'cmd', 'out': 'con'})
#            self.assertEqual(0, r['return'], r.get('error', None))
#            self.assertTrue(r['api'].strip().startswith('Input:'))
#            self.assertTrue(sys.stdout.getvalue().strip().startswith('Function: show cmd'))
#
#        with test_util.tmp_sys():
#            r = ck.get_api({'module_uoa': 'test', 'func': 'cmd', 'out': 'web'})
#            self.assertEqual(0, r['return'], r.get('error', None))
#            self.assertTrue(r['api'].strip().startswith('Input:'))
#            self.assertTrue(sys.stdout.getvalue().strip().startswith('<B>Function:</B> show cmd'))

    def test_parse_cid(self):
        r = ck.parse_cid({'cid': 'a:b:c'})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('a', r['repo_uoa'])
        self.assertEqual('b', r['module_uoa'])
        self.assertEqual('c', r['data_uoa'])

        r = ck.parse_cid({'cid': '#a'})
        self.assertEqual(1, r['return'])
        self.assertEqual('unknown CID format', r['error'])

        r = ck.parse_cid({'cid': '#a', 'ignore_error': 'yes'})
        self.assertEqual(0, r['return'], r.get('error', None))

        r = ck.parse_cid(
            {'cid': '', 'cur_cid': {'repo_uoa': 'r', 'module_uoa': 'm', 'data_uoa': 'd'}})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('r', r['repo_uoa'])
        self.assertEqual('m', r['module_uoa'])
        self.assertEqual('d', r['data_uoa'])

        r = ck.parse_cid(
            {'cid': 'a', 'cur_cid': {'repo_uoa': 'r', 'module_uoa': 'm', 'data_uoa': 'd'}})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('r', r['repo_uoa'])
        self.assertEqual('m', r['module_uoa'])
        self.assertEqual('a', r['data_uoa'])

        r = ck.parse_cid({'cid': 'a:b', 'cur_cid': {
                         'repo_uoa': 'r', 'module_uoa': 'm', 'data_uoa': 'd'}})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('r', r['repo_uoa'])
        self.assertEqual('a', r['module_uoa'])
        self.assertEqual('b', r['data_uoa'])

        r = ck.parse_cid({'cid': 'a:b:c:d', 'cur_cid': {
                         'repo_uoa': 'r', 'module_uoa': 'm', 'data_uoa': 'd'}})
        self.assertEqual(1, r['return'])
        self.assertEqual('unknown CID format', r['error'])

    def test_delete_directory(self):
        with test_util.tmp_dir() as fname:
            os.makedirs(fname + '/a')
            os.makedirs(fname + '/b')

            with open(fname + '/a/test.txt', 'a') as f:
                f.write('abc')

            with open(fname + '/b/test.log', 'a') as f:
                f.write('12')

            r = ck.delete_directory({'path': fname})
            self.assertEqual(0, r['return'], r.get('error', None))
            self.assertFalse(os.path.isdir(fname))

    def test_get_by_flat_key(self):
        a = {'dyn_features': {'ft1': '1', 'ft2': '2'},
             'static_features': {'ft3': '3', 'ft4': '4'}}
        r = ck.get_by_flat_key({'dict': a, 'key': '##dyn_features#ft2'})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('2', r['value'])

    def test_detect_cid_in_current_path(self):
        r = ck.find_path_to_repo({'repo_uoa': 'default'})
        r = ck.detect_cid_in_current_path({'path': r['path']})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('default', r['repo_uoa'])
        self.assertEqual('default', r['repo_alias'])

    def test_uid(self):
        r = ck.uid({})
        self.assertEqual(0, r['return'], r.get('error', None))

        uid = r['data_uid']
        self.assertEqual(16, len(uid))

        chars = '0123456789abcdef'
        for c in uid:
            self.assertIn(c, chars)

    def test_version(self):
        r = ck.get_version({})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual(ck.__version__, r['version_str'])
        self.assertEqual(ck.__version__.split('.'), r['version'])

    def test_check_version(self):
        r = ck.check_version({'version': '1.7.4'})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('yes', r['ok'])
        self.assertEqual(ck.__version__, r['current_version'])

    def test_convert_entry_to_cid(self):
        r = ck.convert_entry_to_cid(
            {'repo_uoa': 'a', 'module_uoa': 'b', 'data_uoa': 'c'})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('a:b:c', r['xcuoa'])

    def test_help(self):
        r = ck.help({})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertTrue('Usage:' in r['help'])
        self.assertTrue('Common actions for' in r['help'])
        for ca in ck.cfg['common_actions']:
            self.assertTrue(ca in r['help'])

    def test_print_input(self):
        d = {'a': 1, 'b': [2, 3]}
        r = ck.print_input(d)
        self.assertEqual(0, r['return'], r.get('error', None))
        parsed_dict = ck.convert_json_str_to_dict(
            {'str': r['html'], 'skip_quote_replacement': 'yes'})['dict']
        self.assertEqual(d, parsed_dict)

    def test_info(self):
        r = ck.info({'module_uoa': 'kernel'})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('kernel', r['data_uoa'])
        self.assertEqual('kernel', r['data_alias'])

    def test_path(self):
        r = ck.find_path_to_repo({'repo_uoa': 'default'})
        r = ck.path({'path': r['path']})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('default', r['repo_uoa'])
        self.assertEqual('default', r['repo_alias'])

    def test_find(self):
        r = ck.find({'module_uoa': 'kernel', 'data_uoa': 'default'})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual(1, r['number_of_entries'])
        self.assertEqual('kernel', r['module_uoa'])
        self.assertEqual('kernel', r['module_alias'])
        self.assertEqual('default', r['data_uoa'])
        self.assertEqual('local', r['repo_alias'])
        self.assertEqual('local', r['repo_uoa'])
        self.assertIn('repo_uid', r)
        self.assertIn('module_uid', r)
        self.assertIn('data_uid', r)
        self.assertIn('path_repo', r)
        self.assertIn('path', r)

#    TBD: cd now opens new bash so need to check how to handle that
#         need to add exit command ...
#
#    def test_cd(self):
#        r = ck.get_os_ck({})
#        plat = r['platform']
#
#        with test_util.tmp_sys():
#            r = ck.cd({'module_uoa': 'kernel', 'data_uoa': 'default'})
#            self.assertEqual(0, r['return'], r.get('error', None))
#
#            if plat == 'win':
#                self.assertEqual('cd /D ' + r['path'], r['string'])
#            else:
#                self.assertEqual('cd ' + r['path'], r['string'])
#
#            # check for fields from ck.find:
#            self.assertEqual(1, r['number_of_entries'])
#            self.assertEqual('kernel', r['module_uoa'])
#            self.assertEqual('kernel', r['module_alias'])
#            self.assertEqual('default', r['data_uoa'])
#            self.assertEqual('local', r['repo_alias'])
#            self.assertEqual('local', r['repo_uoa'])
#            self.assertIn('repo_uid', r)
#            self.assertIn('module_uid', r)
#            self.assertIn('data_uid', r)
#            self.assertIn('path_repo', r)
#            self.assertIn('path', r)

    def test_search(self):
        r = ck.search({'repo_uoa': 'default', 'module_uoa': 'kernel'})
        self.assertEqual(0, r['return'], r.get('error', None))
        lst = r['lst']
        self.assertEqual(1, len(lst))
        r = lst[0]
        self.assertEqual('kernel', r['module_uoa'])
        self.assertEqual('default', r['data_uoa'])
        self.assertEqual('default', r['repo_uoa'])
        self.assertIn('repo_uid', r)
        self.assertIn('module_uid', r)
        self.assertIn('data_uid', r)
        self.assertIn('path', r)

    def test_compare_dicts(self):
        d1 = {'a': 1, 'b': [2, 3]}
        d2 = {'b': [2, 3], 'a': 1}

        r = ck.compare_dicts({'dict1': d1, 'dict2': d2})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('yes', r['equal'])

        d2['b'][1] = 4
        r = ck.compare_dicts({'dict1': d1, 'dict2': d2})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('no', r['equal'])

    def test_compare_flat_dicts(self):
        d1 = {'a': 1, 'b': [2, 3]}
        d2 = {'b': [2, 3], 'a': 1}

        r = ck.compare_flat_dicts({'dict1': d1, 'dict2': d2})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('yes', r['equal'])

        d2['b'][1] = 4
        r = ck.compare_flat_dicts({'dict1': d1, 'dict2': d2})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('no', r['equal'])

    def test_compare_flat_dicts(self):
        d1 = {'a': 1, 'b': [2, 3]}
        d2 = {'b': [2, 3], 'a': 1}

        r = ck.compare_flat_dicts({'dict1': d1, 'dict2': d2})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('yes', r['equal'])

        d2['b'][1] = 4
        r = ck.compare_flat_dicts({'dict1': d1, 'dict2': d2})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('no', r['equal'])

    def test_find_string_in_dict_or_list(self):
        d = {'a': 1, 'b': [2, 3]}

        r = ck.find_string_in_dict_or_list({'dict': d, 'search_string': '2'})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('yes', r['found'])

        r = ck.find_string_in_dict_or_list({'dict': d, 'search_string': '4'})
        self.assertEqual(0, r['return'], r.get('error', None))
        self.assertEqual('no', r['found'])

    def test_list_files(self):
        s = os.sep

        r = ck.list_files(
            {'repo_uoa': 'default', 'module_uoa': 'module', 'data_uoa': 'kernel'})
        self.assertEqual(0, r['return'], r.get('error', None))
        lst = r['list']
        self.assertIn('module.py', lst)
        self.assertIn('test' + s + 'test_kernel.py', lst)
        self.assertIn('test' + s + 'test_original_tests.py', lst)



#
# Collective Knowledge (Tensorflow Model configured with Tensorflow Object Detection API)
#
# 
# 
#
# Developer: 
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

def get_model_env(muid):
    menv_request={
        'module_uoa':'env',
        'data_uoa':muid,
        'action':'load'
    }
    menv=ck.access(menv_request)
    
    menv_return=menv.get('return','1')
    if menv_return != 0:
        return {'return':1, 'error':'Cannot load env: '+muid+'. Maybe this model it is not installed.'}
    return menv

def remove_all_files_from_dir(folder_path):
    import os    
    for the_file in os.listdir(folder_path):
        file_path = os.path.join(folder_path, the_file)
        if os.path.isfile(file_path):
            os.remove(file_path)
    return 0
##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# Train model

def train(i):
    """
    Input:  {
              model_uid    - installed model UID(taken from ck show env)
              (retrain)    - removes previous training/evaluation results
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    import json

    muid=i.get('model_uid','')
    if muid == '':
        return {'return':1, 'error':'Wrong usage: call with --model_uid argument'}

    menv=get_model_env(muid)
    menv_name=menv.get('data_name','') 
   
    menv_dict=menv.get('dict','')
    menv_variables=menv_dict.get('env','')    
    mdir=menv_variables.get('CK_ENV_MODEL_TENSORFLOW_API_ROOT','')
    
    if i.get('retrain','')=='yes':
        ck.out('Model will be retrained...')
        mtdir=menv_variables.get('CK_ENV_MODEL_TENSORFLOW_API_MODEL','')+'/train'
        remove_all_files_from_dir(mtdir)
        medir=menv_variables.get('CK_ENV_MODEL_TENSORFLOW_API_MODEL','')+'/eval'
        remove_all_files_from_dir(medir)
    ck.out('Train model: '+menv_name)

    ck.out('')
    ck.out('Command line: ')
    ck.out('')

    pr={"module_uoa": "program", 
        "data_uoa": "tensorflow-api", 
        "action": "load"
    }
    r=ck.access(pr)
    if r.get('return',1) != 0:
        return {'return':1, 'error':'Cannot load program:tensorflow-api'}    
    r_dict=r.get("dict")
        
    ii={ "module_uoa": "program", 
        "data_uoa": "tensorflow-api", 
        "action": "run",
        "cmd_key":"train",
        "deps":{}}
    
    deps=r_dict.get('run_deps')
    ii['deps']=deps
    ii['deps']['tensorflow-api-model']['uoa']=muid
    
    duid = r.get('data_uid')
    ii['data_uid']=duid

    ck.out('For TensorBoard usage run in another terminal $ tensorboard --logdir='+mdir)
    out=ck.access(ii)
    return {'return':0}

##############################################################################
# Evaluate model 

def eval(i):
    """
    Input:  {
              model_uid    - installed model UID(taken from ck show env)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import json
   
    muid=i.get('model_uid','')
    if muid == '':
        return {'return':1, 'error':'Wrong usage: call with --model_uid argument'}

    menv=get_model_env(muid)
    menv_name=menv.get('data_name','')

    menv_dict=menv.get('dict','')
    menv_variables=menv_dict.get('env','')    
    mdir=menv_variables.get('CK_ENV_MODEL_TENSORFLOW_API_ROOT','')    

    ck.out('Evaluate model: '+menv_name)

    ck.out('')
    ck.out('Command line: ')
    ck.out('')

    pr={"module_uoa": "program", 
        "data_uoa": "tensorflow-api", 
        "action": "load"
    }
    r=ck.access(pr)
    if r.get('return',1) != 0:
        return {'return':1, 'error':'Cannot load program:tensorflow-api'}    
    r_dict=r.get("dict")
        
    ii={ "module_uoa": "program", 
        "data_uoa": "tensorflow-api", 
        "action": "run",
        "cmd_key":"evaluation",
        "deps":{}}
    
    deps=r_dict.get('run_deps')
    ii['deps']=deps
    ii['deps']['tensorflow-api-model']['uoa']=muid
    
    duid = r.get('data_uid')
    ii['data_uid']=duid
    
    ck.out('For TensorBoard usage run in another terminal $ tensorboard --logdir='+mdir)
    out=ck.access(ii)
    return {'return':0}




#
# Collective Knowledge (APK entries)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: cTuning foundation, admin@cTuning.org, http://cTuning.org
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings
import os

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# detect installed APKs

def detect(i):
    """
    Input:  {
              (data_uoa) or (name) - get params only for this APK
              (target_os)          - target Android OS (ck search os --tags=android) (default = android-32)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o=i.get('out','')
    oo=''
    if o=='con': oo='con'

    hos=i.get('host_os','')
    tos=i.get('target_os','')
    tdid=i.get('device_id','')
    target=i.get('target','')

    if target=='' and tos=='':
        tos='android-32'

    ii={'action':'shell',
        'module_uoa':cfg['module_deps']['os'],
        'host_os':hos,
        'target_os':tos,
        'device_id':tdid,
        'target':target,
        'split_to_list':'yes',
        'should_be_remote':'yes',
        'cmd':'pm list packages'}
    r=ck.access(ii)
    if r['return']>0: return r

    tosd=r['target_os_dict']

    lst=r['stdout_lst']

    params={}

    name=i.get('name','')
    if name=='':
        name=i.get('data_uoa','')

    iapk=0
    for package in sorted(lst):
        if package.startswith('package:'):
            package=package[8:]

        if (name!='' and package!=name) or package=='':
            continue

        iapk+=1
        if o=='con':
            ck.out(package)

        params[package]={}

        # Get parameters
        ii={'action':'shell',
            'module_uoa':cfg['module_deps']['os'],
            'host_os':hos,
            'target_os':tos,
            'device_id':tdid,
            'target':target,
            'split_to_list':'yes',
            'should_be_remote':'yes',
            'cmd':'dumpsys package '+package}

        r=ck.access(ii)
        if r['return']>0: return r

        ll=r['stdout_lst']

        for q in ll:
            j=q.find('=')
            if j>0:
                j1=q.rfind(' ', 0, j)
                k=q[j1+1:j]
                v=q[j+1:]

                j2=v.find(' targetSdk=')
                if j2>0:
                    vv=v[j2+11:]
                    v=v[:j2]
                    kk='targetSdk'

                    params[package][kk]=vv

                params[package][k]=v

    if name!='':
        if iapk==0:
            return {'return':16, 'error':'APK was not found on the target device', 'target_os_dict':tosd}

        if o=='con':
            ck.out('')
            ck.out('Parameters:')
            ck.out('')

            for k in sorted(params[name]):
                v=params[name][k]
                ck.out('  '+k+' = '+v)

    return {'return':0, 'params':params, 'target_os_dict':tosd}

##############################################################################
# check APK

def install(i):
    """
    Input:  {
              (host_os)
              (target_os)
              (device_id)

              name or data_uoa   - APK name
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os

    o=i.get('out','')
    oo=''
    if o=='con': oo=o

    name=i.get('name','')
    if name=='':
        name=i.get('data_uoa','')

    if name=='':
        return {'return':1, 'error':'APK "name" is not defined'}

    hos=i.get('host_os','')
    tos=i.get('target_os','')
    tdid=i.get('device_id','')
    target=i.get('target','')

    if target=='' and tos=='':
        tos='android19-arm'
        i['target_os']=tos
#        return {'return':1, 'error':'"target_os" or "target" is not specified'}

    xtdid=''
    if tdid!='': xtdid=' -s '+tdid

    rr={'return':0}

    # Detect if APK is installed
    r=detect(i)
    if r['return']>0 and r['return']!=16: return r

    if r['return']==0:
        rr['params']=r['params']

    if r['return']==16:
        # APK is not installed
        tosd=r['target_os_dict']

        abi=tosd.get('abi','')

        if o=='con':
            ck.out('Searching APK for "'+name+'" and ABI="'+abi+'" ...')

        # Check if available in the CK
        r=ck.access({'action':'load',
                     'module_uoa':work['self_module_uid'],
                     'data_uoa':name})
        if r['return']>0 and r['return']!=16: return r

        found=False
        if r['return']==0:
            p=r['path']
            d=r['dict']

            aname=''
            for apk in d.get('apks',[]):
                if abi in apk.get('abis',[]):
                    aname=apk.get('apk_name','')
                    break

            # If the preferred abi didn't match but is 64-bit,
            # look for a 32-bit binary (worst case won't install)
            alt_abi = ''
            if abi in ['arm64-v8a','arm64']:
                alt_abi='armeabi'
            elif abi=='x86-64':
                alt_abi='x86'
            if alt_abi!='':
                for apk in d.get('apks',[]):
                    if alt_abi in apk.get('abis',[]):
                        aname=apk.get('apk_name','')
                        break

            if aname!='':
                pp=os.path.join(p, aname)

                if os.path.isfile(pp):
                    # Trying to install
                    if o=='con':
                        ck.out('  APK found ('+aname+') - trying to install ...')
                        if alt_abi!='':
                            ck.out('  First choice ABI "'+abi+'" not found, using "'+alt_abi+'"')

                    ii={'action':'shell',
                        'module_uoa':cfg['module_deps']['os'],
                        'host_os':hos,
                        'target_os':hos,
                        'cmd':'adb '+xtdid+' install -r -d '+pp,
                        'out':oo}
                    r=ck.access(ii)
                    if r['return']>0: return r

                    rc=r['return_code']

                    if rc>0:
                        return {'return':1, 'error':'command may have failed (return code='+str(rc)+')'}

                    # Detecting params
                    r=detect(i)
                    if r['return']>0 and r['return']!=16: return r

                    if r['return']==0:
                        rr['params']=r['params']

                    found=True

        # If not found
        if not found:
            if o=='con':
                ck.out('')
                ck.out('APK "'+name+'" with abi "'+abi+'" was not found in CK.')
                ck.out('You can download it and then register in the CK via')
                ck.out(' $ ck add apk:{APK name} --path={full path to downloaded APK}')
                ck.out('')

            return {'return':16, 'error':'APK is not installed on target device and was not found in CK'}

    return rr

##############################################################################
# add apk

def add(i):
    """
    Input:  {
              (data_uoa)  - CK entry to add APK (should be official APK name, i.e. openscience.crowdsource.experiments)
              (repo_uoa)  - repo where to add APK

              (abi)       - list of ABI separated by comma (default=armeabi,armeabi-v7a,arm64-v8a)
              (version)   - version
              (versioncode) - versioncode

              (path)      - path to APK on local host (apk_name will be automatically detected)
              (apk_name)  - force APK name
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os
    import shutil

    o=i.get('out','')

    # Check APK name
    apk_name=i.get('apk_name','')
    path=i.get('path','')
    if path!='':
        if not os.path.isfile(path):
            return {'return':1, 'error':'APK is not found ('+path+')'}

        if apk_name=='':
            apk_name=os.path.basename(path)

    # Check ABI
    abi=i.get('abi','')

    if abi=='':
        r=ck.inp({'text':'Enter list of ABI separated by comma or Enter for "armeabi,armeabi-v7a,arm64-v8a"): '})
        if r['return']>0: return r
        abi=r['string'].strip()

        if abi=='':
            abi='armeabi,armeabi-v7a,arm64-v8a'

    if abi=='':
        return {'return':1, 'error':'"abi" is not specified'}

    abis=abi.split(',')

    # Version
    version=i.get('version','')

    if version=='':
        r=ck.inp({'text':'Enter APK version: '})
        if r['return']>0: return r
        version=r['string'].strip()

    if version=='':
        return {'return':1, 'error':'"version" is not specified'}

    # VersionCode
    versioncode=i.get('versioncode','')

    # Check CK entry name
    duoa=i.get('data_uoa','')
    ruoa=i.get('repo_uoa','')

    if duoa=='':
        r=ck.inp({'text':'Enter CK entry name (must be official APK name): '})
        if r['return']>0: return r
        duoa=r['string'].strip()

    # Check if already exists
    r=ck.access({'action':'load',
                 'module_uoa':work['self_module_uid'],
                 'data_uoa':duoa})
    if r['return']>0 and r['return']!=16: return r

    if r['return']==0:
        ruoa=r['repo_uid']
        pp=r['path']
        dd=r['dict']
    else:
        r=ck.access({'action':'add',
                     'module_uoa':work['self_module_uid'],
                     'common_func':'yes',
                     'data_uoa':duoa,
                     'repo_uoa':ruoa})
        if r['return']>0: return r
        pp=r['path']
        dd={}

    # Create dirs and copy files
    p2=os.path.join(pp,apk_name)

    shutil.copyfile(path, p2)

    # Update dict
    if 'apks' not in dd: dd['apks']=[]
    dd['apks'].append({'abis':abis,
                       'apk_name':apk_name,
                       'version':version,
                       'versioncode':versioncode})

    r=ck.access({'action':'update',
                 'module_uoa':work['self_module_uid'],
                 'data_uoa':duoa,
                 'repo_uoa':ruoa,
                 'dict':dd,
                 'sort_keys':'yes',
                 'substitute':'yes',
                 'ignore_update':'yes'})
    if r['return']>0: return r
    p=r['path']

    if o=='con':
        ck.out('')
        ck.out('APK successfully registered in the CK ('+p+')')

    return r

##############################################################################
# uninstall APK

def uninstall(i):
    """
    Input:  {
              (data_uoa) or (name) - get params only for this APK
              (target_os)          - target Android OS (ck search os --tags=android) (default = android-32)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    # First check if exists
    o=i.get('out','')
    oo=''
    if o=='con':
        i['out']=''
        oo=o

    r=detect(i)
    if r['return']>0: return r

    hos=i.get('host_os','')
    tos=i.get('target_os','')
    tdid=i.get('device_id','')
    target=i.get('target','')

    xtdid=''
    if tdid!='': xtdid=' -s '+tdid

    if target=='' and tos=='':
        tos='android-32'

    name=i.get('name','')
    if name=='':
        name=i.get('data_uoa','')

    ii={'action':'shell',
        'module_uoa':cfg['module_deps']['os'],
        'host_os':hos,
        'target_os':hos,
        'out':oo,
        'cmd':'adb '+xtdid+' uninstall '+name}
    r=ck.access(ii)
    if r['return']>0: return r

    rc=r['return_code']
    if rc>0:
        return {'return':1, 'error':'command may have failed (return code='+str(rc)+')'}

    return r

##############################################################################
# List the apks installed on a target device

def list_installed(i):
    """
    Input:  {
              (device_id)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    # First check if exists
    o=i.get('out','')
    oo=''
    if o=='con':
        i['out']=''
        oo=o

    # r=detect(i)
    # if r['return']>0: return r

    hos=i.get('host_os','')
    tos=i.get('target_os','')
    tdid=i.get('device_id','')
    target=i.get('target','')

    xtdid=''
    if tdid!='': xtdid=' -s '+tdid

    if target=='' and tos=='':
        tos='android-32'

    ii={'action':'shell',
        'module_uoa':cfg['module_deps']['os'],
        'host_os':hos,
        'target_os':hos,
        'out':oo,
        'cmd':'adb '+xtdid+' shell pm list packages -f'}
    r=ck.access(ii)
    if r['return']>0: return r

    rc=r['return_code']
    if rc>0:
        return {'return':1, 'error':'command may have failed (return code='+str(rc)+')'}

    output = r['stdout']

    # Output format is "package:[path]=[package]"
    packages = [ a.split('=')[1] for a in output.split('\n') if '=' in a ]

    if o=='con':
        for p in packages:
            ck.out(p)
    return { 'return':0, 'lst':packages}

##############################################################################
# Uninstall all applications on a target device

def uninstall_all(i):
    """
    Input:  {
              (device_id)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    Uninstall all applications on the device specified.

    """

    # First check if exists
    o=i.get('out','')
    oo=''
    if o=='con':
        i['out']=''
        oo=o

    hos=i.get('host_os','')
    tos=i.get('target_os','')
    tdid=i.get('device_id','')
    target=i.get('target','')

    r=list_installed({'device_id':tdid})
    if r['return']>0: return r

    for apk in r['lst']:
        ii={'data_uoa':apk,
            'device_id':tdid}
        uninstall(ii)
        if r['return']>0: return r

    return {'return':0}



#
# Collective Knowledge (pipeline)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings
sep='***************************************************************************************'
sep1='---------------------------------------------------------------------------------------'

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# setup pipeline

def setup(i):
    """
    Input:  {
              data_uoa     - pipeline module UOA (such as 'benchmark')
                or
              pipeline_uoa
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    o=i.get('out','')

    puoa=i.get('data_uoa','')
    if puoa=='':
       puoa=i.get('pipeline_uoa','')
    if puoa=='':
       return {'return':1, 'error':'pipeline_uoa is not defined'}

    r={'return':0}

    setup_completed=False
    while not setup_completed:
       ii={'action':'pipeline',
           'module_uoa':puoa,
           'prepare':'yes',
           'out':o}
       r=ck.access(ii)
       if r['return']>0: return r

       ready=r.get('ready','')
       if ready=='yes':
          setup_completed=True

    return r

##############################################################################
# universal pipeline autotuning

def autotune(i):
    """
    Input:  {
               (data_uoa)             - pipeline module UOA

               (pipeline)             - prepared pipeline setup (already ready to run)
                     or 
               (pipeline_from_file)   - load prepared pipeline setup from file

               (pipeline_update)      - update pipeline with this dict (useful to update already prepared pipeline from file)

               (pipeline_flags)       - update pipeline directly from the CMD flags (will be parsed by CK)

               (iterations)           - limit number of iterations, otherwise infinite (default=10)
                                        if -1, infinite (or until all choices are explored)
               (start_from_iteration) - skip all iterations before this number
               (repetitions)          - statistical repetitions (default=4)

               (seed)                 - if !='', use as random seed (to reproduce experiments)

               Enforce exploration:
               (start)
               (stop)
               (step)
               (explore_type)         = random, parallel-random, loop, parallel-loop 
                                        Note: machine learning-based or customized autotuners 
                                        are now moved to external plugins 
                                        (see "custom_autotuner" vars below)

                    or
               (random)
               (parallel-random)
               (loop)
               (parallel-loop)

               (process_multi_keys)               - list of keys (starts with) to perform stat analysis on flat array,
                                                       by default ['##characteristics#*', '##features#*' '##choices#*'],
                                                       if empty, no stat analysis

               (record)                           - if 'yes', record results
               (record_uoa)                       - (data UOA or CID where module_uoa ignored!) explicitly record to this entry
               (record_repo)                      - (repo UOA) explicitly select this repo to record
               (record_experiment_repo)           - (repo UOA) explicitly select this repo to record
               (record_failed)                    - if 'yes', record even failed experiments
                                                    (for debugging, buildbots, detecting designed 
                                                     architecture failures, etc)
               (record_only_failed)               - if 'yes', record only failed experiments
                                                    (useful to crowdsource experiments when searching only 
                                                     for compiler/program/architecture bugs  
                                                     (for example fuzzing via random compiler flags))...
               (record_permanent)                 - if 'yes', mark recorded points as permanent (will not be deleted by Pareto filter)
               (record_ignore_update)             - (default=yes), if 'yes', skip recording date/author info for each update
               (tags)                             - record these tags to the entry description
               (subtags)                          - record these subtags to the point description

               (skip_record_pipeline)             - if 'yes', do not record pipeline (to avoid saving too much stuff during crowd-tuning)
               (skip_record_desc)                 - if 'yes', do not record desc (to avoid saving too much stuff during crowd-tuning)

               (record_params)                    - extra record parameters (to 'add experiment' function)

               (features_keys_to_process)         - list of keys for features (and choices) to process/search 
                                                         when recording experimental results (can be wildcards)
                                                         by default ['##features#*', '##choices#*', '##choices_order#*']

               (frontier_keys)                    - list of keys to leave only best points during multi-objective autotuning
                                                     (multi-objective optimization)

               (frontier_keys_reverse)            - list of values associated with above keys. If True, reverse sorting for a give key
                                                    (by default descending)

               (frontier_margins)                 - list of margins when comparing values, i.e. Vold/Vnew < this number (such as 1.10 instead of 1).
                                                    will be used if !=None  

               (frontier_features_keys_to_ignore) - list of keys to ignore from 'features_keys_to_process' 
                                                    when detecting subset of points to detect frontier
                                                    (usually removing optimization dimensions, such as compiler flags)

               (only_filter)                      - if 'yes', do not run pipeline, but run filters on data (for Pareto, for example)

               (skip_stat_analysis)               - if 'yes', just flatten array and add #min

               (stat_flat_dict)                   - pre-load flat dict from previous experiments to aggregate for stat analysis

               (features)                         - extra features
               (meta)                             - extra meta

               (record_dict)                      - extra dict when recording experiments (useful to set subview_uoa, for example)

               (state)                            - pre-load state preserved across iterations

               (save_to_file)                     - if !='', save output dictionary to this file

               (skip_done)                        - if 'yes', do not print 'done' at the end of autotuning

               (sleep)                            - set sleep before iterations ...

               (force_pipeline_update)            - if 'yes', re-check pipeline preparation - 
                                                    useful for replay not to ask for choices between statistical repetitions

               (ask_enter_after_each_iteration)   - if 'yes', ask to press Enter after each iteration

               (tmp_dir)                          - (default 'tmp') - if !='', use this tmp directory to clean, compile and run

               (flat_dict_for_improvements)       - add dict from previous experiment to compare improvements 

               (pause_if_fail)                    - if pipeline fails, ask to press Enter
                                                    (useful to analyze which flags fail during compiler flag autotuning)

               (pause)                            - if 'yes', pause after each iteration

               (aggregate_failed_cases)           - if pipeline fails, aggregate failed cases (to produce report 
                                                    during crowdtuning or automatic compiler bug detection)

               (solutions)                        - check solutions
               (ref_solution)                     - if 'yes', choose reference solution from above list
               (internal_ref)                     - if 'yes', try internal reference before checking solutions ...

               (prune)                            - prune solution (find minimal choices that give the same result)
               (reduce)                           - the same as above
               (reduce_bug)                       - reduce choices to localize bug (pipeline fail)
               (prune_ignore_choices)             - list of choices to ignore (such as base flag, for example)
               (prune_md5)                        - if 'yes', check if MD5 doesn't change
               (prune_invert)                     - if 'yes', prune all (switch off even unused - useful for collaborative machine learning)
               (prune_invert_add_iters)           - if 'yes', add extra needed iterations
               (prune_invert_do_not_remove_key)   - if 'yes', keep both on and off keys (to know exact solution)
               (prune_result_conditions)          - list of extra conditions to accept result (variation, performance/energy/code size constraints, etc)

               (print_keys_after_each_iteration)  - print values of keys from flat dict after each iteration (to monitor characteristics)

               (result_conditions)                - check results for condition
               (condition_objective)              - how to check results (#min,#max,#exp ...)

               (collect_all)                      - collect all experiments and record to 

               (custom_autotuner)                 - dictionary to customize autotuner (exploration, DSE, machine learning based tuning, etc)
               (custom_autotuner_vars)            - extra vars to customize autotuner (for example, set default vs. random)

               (preserve_deps_after_first_run)    - if 'yes', save deps after first run (useful for replay)
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              last_iteration_output - output of last iteration
              last_stat_analysis    - flat dict with stat analysis
              experiment_desc       - dict with experiment description

              recorded_info         - {'points':{}, 
                                       'deleted_points':{},
                                       'recorded_uid'}

              (failed_cases)        - failed cases if aggregate_failed_cases=yes

              (solutions)           - updated solutions with reactions to optimizations (needed for classification of a given computing species)

              (all)                 - if i['collect_all']=='yes', list of results of all iterations [{experiment_desc ..},...]
            }

    """

    import os
    import copy
    import fnmatch
    import json
    import time
    from random import Random

    o=i.get('out','')
    oo=''
    if o=='con': oo='con'

    tmp_dir=i.get('tmp_dir','')

    ic=copy.deepcopy(i)
    ic['module_uoa']=''
    ic['action']=''
    ic['cid']=''
    ic['data_uoa']=''

    pifail=ck.get_from_dicts(ic, 'pause_if_fail', '', None)

    afc=ck.get_from_dicts(ic, 'aggregate_failed_cases', '', None)

    failed_cases=[]

    sols=ck.get_from_dicts(ic, 'solutions', [], None) # Check existing solutions
    isols=len(sols)
    rs=ck.get_from_dicts(ic, 'ref_solution', '', None) 
    iref=ck.get_from_dicts(ic, 'internal_ref', '', None) 

    pdafr=ck.get_from_dicts(ic, 'preserve_deps_after_first_run', '', None)

    prune=ck.get_from_dicts(ic, 'prune', '', None) # Prune existing solutions

    reduce=ck.get_from_dicts(ic, 'reduce', '', None) # Prune existing solutions
    if prune=='': prune=reduce

    reduce_bug=ck.get_from_dicts(ic, 'reduce_bug', '', None) # if 'yes', reduce choices to localize bugs (pipeline fail)

    prune_md5=ck.get_from_dicts(ic, 'prune_md5', '', None) # if 'yes', prune if MD5 doesn't change
    prune_ignore_choices=ck.get_from_dicts(ic, 'prune_ignore_choices', [], None) # Prune existing solutions
    prune_invert=ck.get_from_dicts(ic, 'prune_invert', [], None) # Prune existing solutions
    prune_invert_add_iters=ck.get_from_dicts(ic, 'prune_invert_add_iters', '', None)
    prune_result_conditions=ck.get_from_dicts(ic, 'prune_result_conditions', [], None)
    prune_invert_do_not_remove_key=ck.get_from_dicts(ic, 'prune_invert_do_not_remove_key', [], None)

    print_keys_after_each_iteration=ck.get_from_dicts(ic, 'print_keys_after_each_iteration', [], None)
    lprint_keys_after_each_iteration=-1
    if len(print_keys_after_each_iteration)>0:
       for k in print_keys_after_each_iteration:
           if lprint_keys_after_each_iteration==-1 or len(k)>lprint_keys_after_each_iteration:
              lprint_keys_after_each_iteration=len(k)

    number_of_original_choices=0
    started_prune_invert=False

    condition_objective=ck.get_from_dicts(ic, 'condition_objective', '', None)
    result_conditions=ck.get_from_dicts(ic, 'result_conditions', [], None)

    dsleep=3
    if i.get('sleep','')!='':
       dsleep=float(i['sleep'])

    tags=ck.get_from_dicts(ic, 'tags', [], None)
    tags=ck.convert_str_tags_to_list(tags) # if string, convert to list
    subtags=ck.get_from_dicts(ic, 'subtags', [], None)
    subtags=ck.convert_str_tags_to_list(subtags) # if string, convert to list

    stf=ck.get_from_dicts(ic, 'save_to_file', '', None)

    meta=ck.get_from_dicts(ic, 'meta', {}, None)

    pmk=ck.get_from_dicts(ic, 'process_multi_keys', '', None) # important, default should be '' to be processed later correctly ...

    fk=ck.get_from_dicts(ic, 'frontier_keys', [], None)
    fkr=ck.get_from_dicts(ic, 'frontier_keys_reverse', [], None)
    fmar=ck.get_from_dicts(ic, 'frontier_margins', [], None)

    fdfi=ck.get_from_dicts(ic, 'flat_dict_for_improvements', {}, None)

    record=ck.get_from_dicts(ic, 'record', '', None)
#    We experienced problems with this patch - the problem is somewhere else ...
#    if record=='yes' and len(meta)==0:
#       return {'return':1, 'error':'meta is not defined - can\'t aggregate'}

    record_uoa=ck.get_from_dicts(ic, 'record_uoa', '', None)
    record_repo=ck.get_from_dicts(ic, 'record_repo', '', None)
    record_experiment_repo=ck.get_from_dicts(ic, 'record_experiment_repo', '', None)
    if record.find(':')>=0:
       rx=ck.parse_cid({'cid':record})
       if rx['return']>0: return rx
       record=rx['data_uoa']
       record_repo=rx.get('repo_uoa','')
    record_failed=ck.get_from_dicts(ic, 'record_failed','', None)
    record_only_failed=ck.get_from_dicts(ic, 'record_only_failed','', None)
    record_ignore_update=ic.get('record_ignore_update','')
    if record_ignore_update=='': record_ignore_update='yes'
    if 'record_ignore_update' in ic: del(ic['record_ignore_update'])

    srp=ck.get_from_dicts(ic, 'skip_record_pipeline','', None)
    srd=ck.get_from_dicts(ic, 'skip_record_desc','', None)

    record_permanent=ck.get_from_dicts(ic, 'record_permanent', '', None)

    # Check if repo remote (to save in json rather than to out)
    remote='no'
    if record_repo!='':
       rx=ck.load_repo_info_from_cache({'repo_uoa':record_repo})
       if rx['return']>0: return rx
       remote=rx.get('dict',{}).get('remote','')

    rdict=ck.get_from_dicts(ic, 'record_params', {}, None)
    fkp=ck.get_from_dicts(ic, 'features_keys_to_process', [], None)
    ffki=ck.get_from_dicts(ic, 'frontier_features_keys_to_ignore', [], None)

    state=i.get('state',{})

    jtype=i.get('explore_type','')
    if i.get('random','')=='yes': jtype='random'
    elif i.get('loop','')=='yes': jtype='loop'
    elif i.get('parallel-loop','')=='yes': jtype='parallel-loop'
    elif i.get('parallel-random','')=='yes': jtype='parallel-random'
    cexp={'type':jtype, 
          'omit_probability':i.get('omit_probability',''),
          'start':i.get('start',''),
          'stop':i.get('stop',''),
          'step':i.get('step','')}

    # Check data_uoa
    puoa=i.get('data_uoa','')
    puid=''
    if puoa=='':
       # Try to get from current CID (as module!)
       r=ck.cid({})
       if r['return']==0:
          xruoa=r.get('repo_uoa','')
          xmuoa=r.get('module_uoa','')
          xduoa=r.get('data_uoa','')

          puoa=xmuoa

    if puoa=='':
       return {'return':1, 'error':'data_uoa is not set, i.e. no pipeline module'}

    # Check that data_uoa is module that exists
    r=ck.access({'action':'load',
                 'module_uoa':cfg['module_deps']['module'],
                 'data_uoa':puoa})
    if r['return']>0: return r
    d=r['dict']
    if d.get('pipeline','')!='yes':
       return {'return':1, 'error':'selected pipeline module is not pipeline'}
    puid=r['data_uid']

    # Check meta
    pipeline=i.get('pipeline',{})

    pff=i.get('pipeline_from_file','')
    if pff!='':
       r=ck.load_json_file({'json_file':pff})
       if r['return']>0: return r
       pipeline=r['dict']

    pipeline_update=i.get('pipeline_update',{})

    force_pipeline_update=False
    if len(pipeline_update)!=0:
       r=ck.merge_dicts({'dict1':pipeline, 'dict2':pipeline_update})
       if r['return']>0: return r
       pipeline=r['dict1']

       # Force pipeline update (for example, since changing dataset may change available files)
       force_pipeline_update=True

    pipeline['tmp_dir']=tmp_dir

    # Check if pipeline flags (to update pipeline directly)
    pipeline_flags=i.get('pipeline_flags','')
    if pipeline_flags!='':
       import shlex
       lpipeline_flags=shlex.split('dummy '+pipeline_flags)

       rr=ck.convert_ck_list_to_dict(lpipeline_flags)
       if rr['return']==0:
          dpipeline_flags=rr.get('ck_dict',{})
          if 'cids' in dpipeline_flags: del(dpipeline_flags['cids'])
          if 'action' in dpipeline_flags: del(dpipeline_flags['action'])

          pipeline.update(dpipeline_flags)

    call=ck.get_from_dicts(ic, 'collect_all', {}, None) # Collect all experiemnts
    ae=[]

    # Check customized autotuner
    cat=ck.get_from_dicts(ic, 'custom_autotuner', {}, None) 
    cats=None # script

    if len(cat)>0:
       cmuoa=cat.get('module_uoa','')
       if cmuoa=='':
          x=cat.get('module_uoa_from_pipeline_key','')
          if x!='':
             rx=ck.get_by_flat_key({'dict':pipeline, 'key':x})
             if rx['return']>0: return rx
             cmuoa=rx['value']

       cduoa=cat.get('data_uoa','')
       if cduoa=='':
          x=cat.get('data_uoa_from_pipeline_key','')
          if x!='':
             rx=ck.get_by_flat_key({'dict':pipeline, 'key':x})
             if rx['return']>0: return rx
             cduoa=rx['value']

       try:
           pcat=os.getcwd()
       except OSError:
           os.chdir('..')
           pcat=os.getcwd()

       if cmuoa!='' and cduoa!='':
          rx=ck.access({'action':'find',
                        'module_uoa':cmuoa,
                        'data_uoa':cduoa})
          if rx['return']>0: return rx
          pcat=rx['path']

       cscr=cat.get('script','')

       # Check if has custom script
       rx=ck.load_module_from_path({'path':pcat, 'module_code_name':cscr, 'skip_init':'yes'})
       if rx['return']==0: 
          if 'make' in dir(rx['code']):
             cats=rx['code']
          else:
             return {'return':1, 'error':'autotuner plugin does not have function "make" to make choices'}
       else:
          return rx

    # If pipeline meta is not defined, set up pipeline ...
    fpu=i.get('force_pipeline_update','')

    if len(pipeline)==0 or force_pipeline_update or fpu=='yes':
       if force_pipeline_update or fpu=='yes':
          ii=copy.deepcopy(pipeline)
       else:
          ii=copy.deepcopy(ic)

       ii['out']=o # My latest change (to be able to use local env)
       ii['module_uoa']=puoa
       ii['action']='pipeline'
       ii['prepare']='yes'

       pipeline=ck.access(ii)

       if pipeline['return']>0: return pipeline
       if pipeline.get('fail','')=='yes':
          return {'return':1, 'error':'pipeline setup failed'}
       if pipeline.get('ready','')!='yes':
          return {'return':1, 'error':'pipeline is not ready'}
       del(pipeline['return'])

    # Clean and copy pipeline before choice selection
    for q in cfg['clean_pipeline']:
        if q in pipeline: del(pipeline[q])
    pipelinec=copy.deepcopy(pipeline)

    # Check some vars ...
    ni=i.get('iterations','')
    if ni=='': ni=4
    try: ni=int(ni)
    except Exception as e: pass

    only_filter=ck.get_from_dicts(ic, 'only_filter', '', None)
    if only_filter=='yes':
       ni=1

    ssa=ck.get_from_dicts(ic, 'skip_stat_analysis', '', None)
    sfd=ck.get_from_dicts(ic, 'stat_flat_dict', {}, None)

    sfi=i.get('start_from_iteration','')
    if sfi=='': sfi=1
    if type(sfi)!=int: sfi=int(sfi)

    srm=i.get('repetitions','')
    if srm=='': srm=4

    try: srm=int(srm)
    except Exception as e: pass

    # Check choices descriptions and dimensions
    cdesc=pipeline.get('choices_desc',{})
    corder=copy.deepcopy(i.get('choices_order',[]))

    csel=i.get('choices_selection',{})
    ccur=[]
    pccur={} # pruned choices, if needed

    ref_stat_dict={} # if check with reference, save it
    ref_stat_out={}  # last iteration output for the reference
    ref_rrr={}       # ref characteristics
    ref_rr={}        # ref last iteration output
    keys=[]          # keys to check for conditions or during pruning

    # most likely just one iteration - then set defaults to generate random solutions
    if len(corder)==1 and len(csel)==0:
       csel=[{"type":"random"}]

    # Prepare multi-dimensional vector of choices
    dv1=[] # Current dimensions
    for iq1 in range(0,len(corder)):
        dv=[]
        q1=corder[iq1]
        zz=csel[iq1]
        ztags=[]
        if zz.get('tags','')!='': ztags=zz['tags'].split(',')
        znotags=[]
        if zz.get('notags','')!='': znotags=zz['notags'].split(',')
        zanytags=[]
        if zz.get('anytags','')!='': zanytags=zz['anytags'].split(',')
        for q2 in q1:
            if '*' in q2 or '?' in q2:
               for k in sorted(cdesc, key=lambda v: cdesc[v].get('sort',0)):
                   if fnmatch.fnmatch(k,q2):
                      # Check tags (must have)
                      yy=cdesc[k].get('tags',[])
                      add=True
                      for j in ztags:
                          j=j.strip()
                          if j!='' and j not in yy:
                             add=False
                             break
                      if add:
                         # Any of those
                         if len(zanytags)>0:
                            add=False
                            for j in zanytags:
                                j=j.strip()
                                if j!='' and j in yy:
                                   add=True
                                   break
                         if add:
                            # Check that none of those
                            for j in znotags:
                                j=j.strip()
                                if j!='' and j in yy:
                                   add=False
                                   break
                            if add:
                               dv.append(k)
            else:
               dv.append(q2)
        dv1.append(dv)
    corder=dv1

    # Save original
    ocorder=copy.deepcopy(corder)

    # Check seed
    seed=i.get('seed','')
    if seed=='':
       my_random=Random()
    else:
       my_random=Random(int(seed))

       if o=='con':
          ck.out('')
          ck.out('Random seed: '+str(seed))
          ck.out('')

    points={} # Keep best points if finding best points (possibly with Pareto)
    ppoint={} # Permanent (default point) to calculate improvements ...

    # Start iterations
    rr={'return':0}
    stat_dict={'return':0}
    rrr={'return':0}
    dd={}
    finish=False
    all_solutions=False

    removing_key=''
    removing_value=''
    default_removing_value=''
    cur_md5=''
    last_md5=''
    last_md5_fail_text='compilation returned object file with the same MD5'
    pruned_influence={}
    pruned_chars=[]
    prune_checked_keys=[]
    pruned_inversed_flags={}
    prune_check_all=[]

    sol={}

    increased_iterations=False

    recorded_info={}
    last_record_uid=''

    ref_keys={}

    m=-1
    while True:
        m+=1
        if ni!=-1 and m>=ni:
           break

        mm=m+1

        x='Pipeline iteration: '+str(mm)
        if ni!=-1: x+=' of '+str(ni) 

        ck.out(sep1)
        ck.out(x)

        time.sleep(0.5)

        # Copy original
        if m==0 or mm>=sfi:
           pipeline=copy.deepcopy(pipelinec)

        # Check if there is a pre-selection
        al=''
        suid=''
        if isols>0:
           if not (iref=='yes' and m==0) and not all_solutions:
              if o=='con':
                 x=''
                 if rs=='yes':
                    x=' reference'

                 ck.out('')
                 ck.out('  Checking pre-existing'+x+' solution ...')

              # Find choices (iterate over solutions since there can be more 
              #  than one point in each solution due to Pareto)

              nz=[] # remaining keys to select from (if 0, stop)

              if rs=='yes':
                 sol=sols[0]
                 corder1=sol.get('ref_choices_order',[])
                 cx1=sol.get('ref_choices',{})
                 suid=sol.get('solution_uid','')
              else:
                 if prune=='yes':
                    if len(pccur)>0:
                       if o=='con':
                          ck.out('')
                          ck.out('  Pruning or inverting solution (iteratively reducing complexity of the found solution) ...')

                       # Check non-zero ones:
                       if not started_prune_invert:
                          for g in range(0, len(corder1)):
                              k=corder1[g]
                              v=pccur.get(k, None)
                              if v!='' and v!=None and k not in prune_ignore_choices and k not in prune_checked_keys:
                                 nz.append(k)

                       # Remove one (random or one by one)
                       if len(nz)==0:
                          if not started_prune_invert:
                             if o=='con':
                                ck.out('')
                                ck.out('  ALL choices have been checked - main pruning finished !')

                          if prune_invert!='yes':
                             break

                          if not started_prune_invert:
                             ck.out('  ***')
                             ck.out('  Starting iteratively inverting all possible choices ...')

                             for q in ocorder[0]:
                                 if q not in corder1:
                                    corder1.append(q)

                             for q in corder1:
                                 if q not in prune_checked_keys and q not in pccur and q in cdesc:
                                    cddx=cdesc[q]
                                    if 'boolean' in cddx.get('tags',[]):
                                       prune_check_all.append(q)

                             started_prune_invert=True

                          if len(prune_check_all)==0:
                             ck.out('')
                             ck.out('  Pruning finished !')

                             break

                          # Switching off key
                          y=my_random.randrange(0,len(prune_check_all))

                          removing_key=prune_check_all[y]
                          del(prune_check_all[y])

                          cddx=cdesc.get(removing_key,{})
                          cccx=cddx.get('choice',[])

                          v=''
                          default_removing_value=''
                          if len(cccx)>0:
                             v=cccx[len(cccx)-1]
                             default_removing_value=cccx[0]

                          removing_value=v
                          pccur[removing_key]=v

                          cx1=copy.deepcopy(pccur)

                          if o=='con':
                             ck.out('')
                             ck.out('    Trying to invert key '+removing_key+' to default ('+v+') ...')

                       else:
                          y=my_random.randrange(0,len(nz))

                          removing_key=nz[y]
                          removing_value=pccur[removing_key]

                          del(nz[y])

                          pccur[removing_key]=''

                          prune_checked_keys.append(removing_key)

                          cx1=copy.deepcopy(pccur)

                          if o=='con':
                             ck.out('')
                             ck.out('    Trying to remove key '+removing_key+' ('+str(removing_value)+') ...')

                    # Increase iterations if check all
                    if len(nz)==0 and prune_invert_add_iters=='yes' and prune_invert=='yes' and not increased_iterations:
                       increased_iterations=True
                       for q in ocorder[0]:
                           if q not in prune_checked_keys and q not in pccur and q in cdesc:
                              cddx=cdesc[q]
                              if 'boolean' in cddx.get('tags',[]):
                                 ni+=1

                 if prune!='yes' or len(pccur)==0:
                    sol={}
                    bb=0
                    if iref=='yes': bb=1
                    for b in sols:
                        bp=b.get('points',[])
                        suid=b.get('solution_uid','')
                        for bx in bp:
                            if m==bb:
                               sol=bx
                               break
                            bb+=1
                        if len(sol)>0:
                           keys=list(sol.get('improvements',{}).keys())
                           break

                    if len(sol)==0:
                       all_solutions=True
                       suid=''
                    else:
                       corder1=sol.get('pruned_choices_order',[])
                       cx1=sol.get('pruned_choices',{})

                    if suid!='' and o=='con':
                       ck.out('     Solution UID: '+suid+' ...')

              if all_solutions:
                 # If checked all solutions, reset original choices and start (random) exploration
                 corder=copy.deepcopy(ocorder)
                 ccur=[]
                 al=''

                 if o=='con':
                    ck.out('')
                    ck.out('  ALL pre-existing solutions validated - restarting extra exploration ...')

              else:
                 corder2=[]
                 ccur2=[]

                 for b in corder1:
                     v=cx1.get(b,None)
                     if v!='' and v!=None:
                        ccur2.append(v)
                        corder2.append(b)

                 # Check if reuse ref choices
                 if sol.get('preset_ref_choices','')=='yes' and len(ref_keys)>0:
                    for k in sorted(ref_keys):
                        if k not in corder2:
                           corder2.append(k)
                           ccur2.append(ref_keys[k])

                 corder=[corder2]
                 ccur=[ccur2]

                 al='yes'

                 if prune=='yes' and len(pccur)==0:
                    pccur=copy.deepcopy(cx1)

        # Make selection
        jj={'module_uoa':cfg['module_deps']['choice'],
            'action':'make',
            'choices_desc':cdesc,
            'choices_order':corder,
            'choices_selection':csel,
            'choices_current':ccur,
            'custom_explore':cexp,
            'pipeline':pipeline,
            'random_module':my_random,
            'out':o}
        if al!='': jj['all']=al

        if cats!=None and al!='yes':
           # customized autotuning via external module (exploration)
           # specialize our autotuner to a given program (domain)

           jj['ck_kernel']=ck
           jj['state']=state
           jj['meta']=meta
           jj['autotuning_iteration']=m
           jj['tmp_dir']=tmp_dir
           jj['custom_autotuner_meta']=cat

           jj['vars']=i.get('custom_autotuner_vars',{})

           r=cats.make(jj)
           if r['return']>0: return r

           # Print keys and update pipeline!
           ks=r.get('keys',{})
           if len(ks)>0:
              if o=='con': 
                 ck.out('')

              for k in sorted(ks):
                  v=ks[k]
                  if o=='con': 
                     ck.out(' * '+k+' = '+str(v))

                  rx=ck.set_by_flat_key({'dict':pipeline, 'key':k, 'value':v})
                  if rx['return']>0: return rx

              if o=='con': 
                 ck.out('')

              # Save ref choices just in case
              if m==0:
                 ref_keys=copy.deepcopy(ks)

        else:
           # CK-based selection 
           r=ck.access(jj)
           if r['return']>0: return r

        if r.get('finish',True):
           finish=True
           break

        if i.get('ask_enter_after_choices','')=='yes':
           ck.out('')
           ck.inp({'text':'Press Enter to continue ...'})

        time.sleep(dsleep) # wait to see selection ...

        # Check if pass this iteration
        if mm<sfi:
           ck.out('')
           ck.out('  (skiped by request)')
           continue

        # Describing experiment
        dd={'tags':tags,
            'subtags':subtags,
            'meta':meta,
            'pipeline':pipelinec,
            'pipeline_uoa':puoa,
            'pipeline_uid':puid}

        ddcl=[] # characteristics list (from statistical repetitions)

        fail='no'
        fail_reason=''
        rr={}
        last_orders_from_pipeline=[]
        for sr in range(0, srm):
            if only_filter=='yes': continue

            ck.out('')
            ck.out('      ------------------- Statistical repetition: '+str(sr+1)+' of '+str(srm)+' -------------------')
            ck.out('')

            pipeline1=copy.deepcopy(pipeline)

            pipeline1['prepare']='no'
            pipeline1['module_uoa']=puoa
            pipeline1['action']='pipeline'
            pipeline1['out']=o
            pipeline1['state']=state
            pipeline1['meta']=meta
            pipeline1['autotuning_iteration']=m
            pipeline1['statistical_repetition_number']=sr
            pipeline1['tmp_dir']=tmp_dir

#            if prune_md5=='yes':

            # If last pruning iteration, compile and run code (to get all last characteristics)
            if prune_md5=='yes' and (len(nz)!=0 or len(prune_check_all)!=0):
               if o=='con' and last_md5!='':
                  ck.out('')
                  ck.out('      Checking MD5: '+last_md5)

               pipeline1['last_md5']=last_md5
               pipeline1['last_md5_fail_text']=last_md5_fail_text

            rr=ck.access(pipeline1)
            if rr['return']>0: return rr

            fail=rr.get('fail','')
            fail_reason=rr.get('fail_reason','')

            if pdafr=='yes' and m==0 and sr==0:
               # Preserve resolved deps (useful for replay)
               pipeline['dependencies']=copy.deepcopy(pipeline1.get('dependencies',{}))
               pipelinec['dependencies']=copy.deepcopy(pipeline1.get('dependencies',{}))

               # Check if reduce bug but pipeline didn't fail
               if reduce_bug=='yes' and fail!='yes':
                  return {'return':1, 'error':'trying to reduce bug but pipeline didn\'t fail'}

            xcur_md5=pipeline1.get('characteristics',{}).get('compile',{}).get('md5_sum','')
            if sr==0 and xcur_md5!='': cur_md5=xcur_md5

            state=rr.get('state',{})

            last_orders_from_pipeline=rr.get('choices_order',[])

            if (fail!='yes' and record_only_failed!='yes') or (fail=='yes' and (record_failed=='yes' or record_only_failed=='yes')):
               ddcl.append(pipeline1.get('characteristics',{}))

               if sr==0: # record for the first iteration otherwise pipeline may change via state ...
                  dd['dict']=i.get('record_dict',{}) # useful for subview_uoa
                  dd['choices']=rr.get('choices',{})
                  dd['choices_order']=rr.get('choices_order',[])
                  dd['choices_desc']=rr.get('choices_desc',{})
                  dd['features']=rr.get('features',{})
                  dd['features']['statistical_repetitions']=srm
                  if suid!='':
                     dd['features']['solution_uid']=suid
                  dd['features_desc']=rr.get('features_desc',{})
                  dd['dependencies']=rr.get('dependencies',{})
                  dd['characteristics_desc']=rr.get('characteristics_desc',{})

                  # TBD: CHECK HOW TO WRITE META HERE DEPENDING ON SCENARIOS ...

                  meta1=rr.get('meta',{})
                  if len(meta1)>0: meta.update(meta1)
                  dd['meta']=meta

            if fail=='yes': break

        # Record extra pipeline info
        fail_bool=False
        if fail=='yes': fail_bool=True
        dd['pipeline_state']={'repetitions':srm,
                              'fail_reason':fail_reason,
                              'fail':fail,
                              'fail_bool':fail_bool}

        # Record list of characteristics (from multiple reptitions)
        dd['characteristics_list']=ddcl

        if afc=='yes' and fail_bool:
           failed_cases.append({'choices':rr.get('choices',{}),
                                'choices_order':rr.get('choices_order',[]),
                                'features':rr.get('features',{}),
                                'characteristics':rr.get('characteristics',{}),
                                'pipeline_state':dd.get('pipeline_state',{})})

        if i.get('pause','')=='yes' or (fail_bool and pifail=='yes'):
           ck.out('')
           ck.inp({'text':'Press Enter to continue ...'})

        ##########################################################################################
        # Collect all results if needed to be processed by user (customized crowd-tuning, for example)
        if call=='yes':
           ae.append(dd)

        ##########################################################################################
        # Recording experiment if needed
        current_point=''
        current_record_uid=''

        stat_dict={}

        recorded_info={'points':[], 'deleted_points':[], 'recorded_uid':''}

        iec={'module_uoa':cfg['module_deps']['experiment'],

             'repo_uoa': record_repo,
             'experiment_repo_uoa': record_experiment_repo,
             'experiment_uoa':record_uoa,

             'record_permanent':record_permanent,

             'features_keys_to_process':fkp}
        fft={}

        if record=='yes' and only_filter!='yes':
           # Will be reused for frontier
           fft={} # flat features

           if (fail!='yes' and record_only_failed!='yes') or (fail=='yes' and (record_failed=='yes' or record_only_failed=='yes')):

              t1=time.time()

              if o=='con':
                 ck.out(sep)
                 ck.out('Recording experiment ...')
                 ck.out('')

              ie=copy.deepcopy(iec)

              if remote=='yes': ie['out']=''
              else: ie['out']=oo

              if len(fdfi)>0: 
                 dd['dict_to_compare']=fdfi

              ie['action']='add'
              ie['ignore_update']=record_ignore_update
              ie['sort_keys']='yes'
              ie['record_all_subpoints']='yes'
              ie['process_multi_keys']=pmk
              ie['dict']=dd

              if srp!='': ie['skip_record_pipeline']=srp
              if srd!='': ie['skip_record_desc']=srd

              if ssa!='': ie['skip_stat_analysis']=ssa

              # Update what we record and how we process data from external dict -> 
              #  may be customized depending on scenarios 
              #  (compiler flag tuning, OpenCL/MPI params, etc)
              ie.update(rdict)

              rx=ck.access(ie)
              if rx['return']>0: return rx

              current_point=rx.get('point','')
              current_record_uid=rx.get('recorded_uid','')

              if current_point!='': recorded_info['points'].append(current_point)
              if current_record_uid!='': 
                 recorded_info['recorded_uid']=current_record_uid
                 last_record_uid=current_record_uid

              stat_dict=rx['dict_flat']
              rrr=rx['stat_analysis']
              fft=rx['flat_features']

              tt=time.time()-t1
              if o=='con':
                 ck.out('')
                 ck.out('Recorded successfully in '+('%.2f'%tt)+' secs.')

        ##########################################################################################
        # If was not performed via recording, perform statistical analysis here
        if ssa!='yes' and fail!='yes' and len(stat_dict)==0:
           if o=='con':
              ck.out('')
              ck.out('Performing explicit statistical analysis of experiments ...')

           if prune=='yes': sfd={}

           ii={'action':'multi_stat_analysis',
               'module_uoa':cfg['module_deps']['experiment'],
               'dict':stat_dict,
               'flat_dict':sfd,
               'dict_to_add':dd,
               'dict_to_compare':fdfi,
               'process_multi_keys':pmk,
               'skip_stat_analysis':ssa}
           rrr=ck.access(ii)
           if rrr['return']>0: return rrr
           stat_dict=rrr['dict_flat']

        ##########################################################################################
        # Print various values to monitor iterations (if needed)
        if o=='con' and fail!='yes' and len(print_keys_after_each_iteration)>0:

           ck.out('')
           ck.out('Monitoring characteristics:')
           ck.out('')

           for k in print_keys_after_each_iteration:
               v=stat_dict.get(k,None)
               x='  * '+(' '*(lprint_keys_after_each_iteration-len(k)))+str(k)+' = '+str(v)

               vr=ref_stat_dict.get(k,None)
               if v!=None and vr!=None:
                  x+=' ('+str(vr)+')'

               ck.out(x)

        ##########################################################################################
        # Check conditions
        if current_point!='' and len(result_conditions)>0 and (prune!='yes' or m>0) and fail!='yes':
           if o=='con':
              ck.out('')
              ck.out('Check extra conditions on results ...')

           # Check conditions
           ii={'action':'check',
               'module_uoa':cfg['module_deps']['math.conditions'],
               'new_points':['0'],
               'results':[{'point_uid':'0', 'flat':stat_dict}],
               'conditions':result_conditions,
               'middle_key':condition_objective,
               'out':oo}
           ry=ck.access(ii)
           if ry['return']>0: return ry 

           xdpoints=ry['points_to_delete']
           if len(xdpoints)>0:
              if o=='con':
                 ck.out('')
                 ck.out('    WARNING: conditions on characteristics were not met!')

              # Delete point and fail
              rx=ck.access({'action':'delete_points',
                            'module_uoa':cfg['module_deps']['experiment'],
                            'points':[{'module_uid':cfg['module_deps']['experiment'], 
                                       'data_uid':current_record_uid,
                                       'point_uid':current_point}],
                            'out':oo})
              if rx['return']>0: return rx

              rr['fail']='yes'
              rr['fail_reason']='conditions were not met'

              fail=rr.get('fail','')
              fail_reason=rr.get('fail_reason','')

        ##########################################################################################
        # If checking pre-recoreded solutions, add result
        if prune!='yes' and isols>0 and m<isols:
           if o=='con':
              ck.out('')
              ck.out('Recording reaction to optimizations to a pre-existing solution (for automatic classification of computational species) ...')
              ck.out('')

           # Prepare improvements/degradations
           reaction=False
           bb=0
           for jb in range(0, len(sols)):
               b=sols[jb]
               bp=b.get('points',[])
               for by in range(0, len(bp)):
                   if m==bb:
                      bp[by]['reaction_raw_flat']=copy.deepcopy(stat_dict)
                      bp[by]['reaction_info']={'fail':fail, 'fail_reason':fail_reason}

                      imp=copy.deepcopy(bp[by].get('improvements',{}))
                      if len(imp)>0:
                         for k in imp:
                             imp[k]=stat_dict.get(k,None)
                         bp[by]['improvements_reaction']=imp

                      reaction=True
                   bb+=1
               if reaction:
                  sols[jb]['points']=bp
                  break

        # Check if need to leave only points on frontier 
        #   (our frontier is not 'Pareto efficient' since we have discreet characteristics)

#        if len(fk)>0 and (len(stat_dict)>0 or only_filter=='yes'): - otherwise doesn't work if last point is error

        # If checking pruning (that results didn't change much)
        if prune=='yes':
           # Rebuild list of current choices from last choices dict (since pipeline may change choices, such as base flag)
           if len(fft)>0:
              pccur={}
              for q in last_orders_from_pipeline:
                  q1='##choices'+q[1:]
                  v=fft.get(q1,'')
                  if v!='' and v!=None:
                     pccur[q]=v

           if m==0:
              # Saving reference stats
              ref_stat_dict=copy.deepcopy(stat_dict)
              ref_stat_out=copy.deepcopy(rr)
              fdfi=ref_stat_dict
              last_md5=cur_md5 #stat_dict.get('##characteristics#compile#md5_sum#min',None)

              for q in pccur:
                  if q!=None and q!='':
                     number_of_original_choices+=1

              if o=='con':
                 ck.out('')
                 ck.out('Preserving reference statistics to compare against during solution pruning ...')
                 ck.out('')

           else:
              if o=='con':
                 ck.out('')
                 ck.out('Checking if results have changed during pruning ...')
                 ck.out('')

                 result_the_same=False
                 if fail=='yes':
                    if reduce_bug=='yes' or (fail_reason==last_md5_fail_text):
                       result_the_same=True

                 if not result_the_same and fail!='yes':
                    if len(prune_result_conditions)==0:
                       result_the_same=True
                    else:
                       ck.out('')
                       ck.out('Check extra pruning conditions on results ...')
                       ck.out('')

                       # Check conditions
                       ii={'action':'check',
                           'module_uoa':cfg['module_deps']['math.conditions'],
                           'new_points':['0'],
                           'results':[{'point_uid':'0', 'flat':stat_dict}],
                           'conditions':prune_result_conditions,
                           'middle_key':condition_objective,
                           'out':oo}
                       ry=ck.access(ii)
                       if ry['return']>0: return ry 

                       gpoints=ry['good_points']
                       dpoints=ry['points_to_delete']
                       pruned_chars=ry['keys']

                       if len(dpoints)==0:
                          result_the_same=True

                 if o=='con':
                    ck.out('')
                    if result_the_same:
                       ck.out('    *** Result did not change!')
                    else:
                       ck.out('    *** Result changed!')

                 restore_stats=False
                 if started_prune_invert:
                    # We are in inverting mode ######################################################
                    if result_the_same:
                       if o=='con':
                          ck.out('')
                          ck.out('    Keeping default key "'+removing_key+'" ...')

                       pccur[removing_key]=removing_value

                       # If new compilation MD5 and new stats, make a new referneces point
                       if not (fail=='yes' and fail_reason==last_md5_fail_text):
                          ref_stat_dict=copy.deepcopy(stat_dict)
                          ref_stat_out=copy.deepcopy(rr)
                          ref_rrr=copy.deepcopy(rrr)
                          ref_rr=copy.deepcopy(rr)
                          fdfi=ref_stat_dict
                          last_md5=cur_md5

                          if o=='con':
                             ck.out('')
                             ck.out('         NEW REFERENCE POINT (MD5='+last_md5+')!')

                    else:
                       if prune_invert_do_not_remove_key=='yes':
                          prx1='Adding'
                          prx1a='to'
                          prx2=default_removing_value
                       else:
                          prx1='Removing'
                          prx1a='from'
                          prx2=''

                       if o=='con':
                          ck.out('')
                          ck.out('    '+prx1+' key "'+removing_key+'" '+prx1a+' choices ...')

                       pccur[removing_key]=prx2

                       pruned_inversed_flags[removing_key]=prx2

                 else:
                    # We are in standard pruning mode ###############################################
                    if result_the_same:
                       if o=='con':
                          ck.out('')
                          ck.out('    Removing key "'+removing_key+'" from choices ...')

                       pccur[removing_key]=''

                       pruned_inversed_flags[removing_key]=removing_value

                       # If new compilation MD5 and new stats, make a new referneces point
                       if not (fail=='yes' and fail_reason==last_md5_fail_text):
                          ref_stat_dict=copy.deepcopy(stat_dict)
                          ref_stat_out=copy.deepcopy(rr)
                          ref_rrr=copy.deepcopy(rrr)
                          ref_rr=copy.deepcopy(rr)
                          fdfi=ref_stat_dict
                          last_md5=cur_md5 #stat_dict.get('##characteristics#compile#md5_sum#min',None)

                          if o=='con':
                             ck.out('')
                             ck.out('         NEW REFERENCE POINT (MD5='+last_md5+')!')

                    else:
                       if o=='con':
                          ck.out('')
                          ck.out('    Keeping key "'+removing_key+'" ...')

                       pccur[removing_key]=removing_value

                       restore_stats=True

                 # Record influential optimization
                 if not result_the_same:
                    kky={}
                    if len(pruned_chars)>0:
                       for q in pruned_chars:
                           kky[q]=stat_dict.get(q,None)
                    pruned_influence[removing_key]=kky

                 # Restore stat dict (for comparison)
                 if restore_stats:
                    stat_dict=copy.deepcopy(ref_stat_dict)
                    rrr=copy.deepcopy(ref_rrr)
                    rr=copy.deepcopy(ref_rr)

        if prune!='yes' and (len(fk)>0 or only_filter=='yes'):
           # If data was recorded to repo, reload all points 
           opoints={} # original points with all info (used later to delete correct points)
           if record=='yes':
              if o=='con':
                 ck.out('')
                 ck.out('Reloading points to detect frontier ...')
                 ck.out('')

              ie=copy.deepcopy(iec)

              ie['action']='get'

              if only_filter=='yes':
                 ie['get_all_points']='yes'
              else:
                 # NOTE - I currently do not search points by features, i.e. all points are taken for Paretto
                 # should improve in the future ...
                 ie['flat_features']=fft
                 ie['features_keys_to_ignore']=ffki # Ignore parts of features to be able to create subset for frontier ...
                 if 'features_keys_to_process' in ie: del(ie['features_keys_to_process'])
                 ie['skip_processing']='yes'

              ie['separate_permanent_points']='yes'
              ie['load_json_files']=['flat','features']
              ie['get_keys_from_json_files']=fk
              ie['out']='con'

              rx=ck.access(ie)
              if rx['return']>0: return rx

              xpoints=rx['points']
              opoints=copy.deepcopy(xpoints) # original points

              points={}

              ppoints=rx['ppoints']
              ppoint={}
              if len(ppoints)>0:
                 ppoint=ppoints[0]

              for q in xpoints:
                  uid=q['point_uid']
                  points[uid]=q.get('flat',{})
           else:
              rx=ck.gen_uid({})
              if rx['return']>0: return rx
              uid=rx['data_uid']

              w={}
              for q in stat_dict: 
                  if q in fk: w[q]=stat_dict[q]

              points[uid]=w

           # Filtering ...
           if o=='con':
              ck.out(sep)
              ck.out('Filtering multiple characteristics on frontier (not necessarily Pareto) for multi-objective autotuning')
              ck.out('')

           # Save originals to know which ones to delete if recording ...
           rx=ck.access({'action':'filter',
                         'module_uoa':cfg['module_deps']['math.frontier'],
                         'points':points,
                         'frontier_keys':fk,
                         'reverse_keys':fkr,
                         'margins':fmar,
                         'out':oo})
           if rx['return']>0: return rx

           points=rx['points']
           dpoints=rx['deleted_points']

           recorded_info['deleted_points']=list(dpoints.keys())

           recorded_info['points']=list(points.keys())
           if len(ppoint)>0:
               recorded_info['points'].append(ppoint['point_uid'])

           # Delete filtered if record
           if record=='yes':
              xdpoints=[]
              for q in dpoints:
                  found=False
                  for qq in opoints:
                      if qq['point_uid']==q:
                         found=True
                         break
                  if found:
                     xdpoints.append(qq)

              # Attempt to delete non-optimal solutions
              rx=ck.access({'action':'delete_points',
                            'module_uoa':cfg['module_deps']['experiment'],
                            'points':xdpoints,
                            'out':oo})
              if rx['return']>0: return rx

        if i.get('ask_enter_after_each_iteration','')=='yes':
           ck.out('')
           ck.inp({'text':'Press Enter to continue autotuning or DSE ...'})

    # Mention, if all iterations were performed, or autotuning rached max number of iterations
    if finish:
       ck.out('')
       ck.out('All iterations are done!')

    recorded_info['last_recorded_uid']=last_record_uid
    rz={'return':0, 'last_iteration_output':rr, 'last_stat_analysis': rrr, 'experiment_desc':dd, 'recorded_info':recorded_info, 'failed_cases':failed_cases}

    if call=='yes':
       rz['all']=ae

    # If pruning, print last results
    report=''
    if prune=='yes' and o=='con':
       rz['last_stat_analysis']=ref_stat_dict
       rz['last_iteration_output']=ref_stat_out

       ll=0
       lx=0
       for k in corder1:
           if len(k)>ll: 
              ll=len(k)
           v=pccur.get(k, None)
           if v!='' and v!=None:
              lx+=1

       x=str(lx)
       if prune_invert!='yes':
          x+=' vs '+str(number_of_original_choices)

       report='\n  Final pruned choices ('+x+') :\n'

       report+='        Characteristics\' changes (in brackets):\n'
       for q in range(0, len(pruned_chars)):
           pc=pruned_chars[q]
           report+='           '+str(q)+' = '+pc+'\n'
       report+='\n'

       keys=[]
       pruned={}
       pruned_order=[]

       for k in corder1:
           v=pccur.get(k,None)
           if v!='' and v!=None:
              pruned_order.append(k)
              pruned[k]=v
              keys.append(k)

       for k in pruned_influence:
           if k not in keys:
              keys.append(k)

       compiler_flags=''

       for k in keys:

           v=pccur.get(k,None)
           v1=pruned_influence.get(k,None)

           if v!=None and k.startswith('##compiler_flags#'): 
              compiler_flags+=' '+v.strip()

           j=ll-len(k)

           x=' '*j

           y=''
           if v1!=None:
              for q in pruned_chars:
                  v1x=v1.get(q,0.0)
                  if v1x==None: v1x=0.0
                  if y!='': y+=';'
                  v2=float(v1x)
                  y+=('%1.3f' % v2)
           else:
              y='           '
           report+='    '+k+x+' ('+y+') : '+str(v)+'\n'

       compiler_flags=compiler_flags.strip()
       if compiler_flags!='':
          report+='\n        Compiler flags: '+compiler_flags+'\n'
          report+='\n'

       if o=='con':
          ck.out(report)

       rz['report']=report

       # Embedd to first solution:
       b=sols[0]['points'][0]

       b['pruned_choices_order']=pruned_order
       b['pruned_choices']=pruned

       b['pruned_influence']=pruned_influence
       b['pruned_chars']=pruned_chars
       b['pruned_inversed_flags']=pruned_inversed_flags

       sols[0]['points'][0]=b

    if m>0 and i.get('skip_done','')!='yes':
       ck.out(sep)
       ck.out('Done!')

    if len(sols)>0:
       rz['solutions']=sols

    if stf!='':
       rx=ck.save_json_to_file({'json_file':stf, 'dict':rz, 'sort_keys':'yes'})

    return rz

##############################################################################
# Run pipeline once ...

def run(i):
    """
    See "autotune" API with iterations=1
    """

    if i.get('iterations','')=='': i['iterations']=1
    return autotune(i)



#
# Collective Knowledge (Check speedup of program versus various compiler flags and data sets)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings
sep='**********************************************************************'

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# describe experiment

def describe(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    ck.out(cfg['full_desc'])

    return {'return':0}

##############################################################################
# reproduce experiment

def reproduce(i):
    """
    Input:  {
              program_uoa       - program UOA to check

              (cmd_key)         - cmd key
              (dataset_uoas)    - check dataset UOA

              (choices)         - dict['flags'] - list of combinations of compiler flags

              (host_os)         - host OS (detect, if omitted)
              (target_os)       - OS module to check (if omitted, analyze host)
              (device_id)       - device id if remote (such as adb)

              (stat_repeat)     - max statistical repetitions (4 by default)

              (check_speedup)   - if 'yes', check speedups for the first two optimizations ...

              (add_to_pipeline) - add this dict to pipeline 

              (force_record)    - if 'yes', force record even if behavior expected ...

              (experiment_repo_uoa)        - repo to record experiments (by default "remote-ck")
              (experiment_remote_repo_uoa) - if above repo is remote, repo on remote server to record experiments (by default "upload")
              (experiment_uoa)             - CK entry UOA to record experiments (by default "reproduce-ck-paper-filter-optimization")
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    import os
    curdir=os.getcwd()
    rf=os.path.join(curdir, cfg['report_file'])

    puoa=i.get('program_uoa','')
    if puoa=='':
       return {'return':1, 'error':'program_uoa is not defined.\n\nUse "ck list program" to see available programs.\nUse "ck pull repo:ck-programs" and "ck pull repo:ck-datasets-min" to get a small set of our benchmarks and datasets.'}

    choices=i.get('choices',{})
    if len(choices)==0:
       choices=cfg['choices']

    cflags=choices.get('flags',[])
    if len(cflags)==0:
       return {'return':1, 'error':'choices dictionary doesn\'t have "flags" list'}

    ap=i.get('add_to_pipeline',{})

    e_repo_uoa=cfg['repository_to_share_results']
    e_remote_repo_uoa=cfg['remote_repo_uoa']
    e_uoa=cfg['remote_experiment_uoa']

    if i.get('experiment_repo_uoa','')!='': e_repo_uoa=i['experiment_repo_uoa']
    if i.get('experiment_remote_repo_uoa','')!='': e_remote_repo_uoa=i['experiment_remote_repo_uoa']
    if i.get('experiment_uoa','')!='': e_uoa=i['experiment_uoa']

    ###################################################
    # Experiment table
    table=[]  # Strings (for printing)
    otable=[] # Original format

    ###################################################
    ck.out(sep)
    ck.out('Loading program meta info ...')

    r=ck.access({'action':'load',
                 'module_uoa':cfg['module_deps']['program'],
                 'data_uoa':puoa})
    if r['return']>0: return r
    pd=r['dict']

    cmd_key=i.get('cmd_key','')
    if cmd_key=='': cmd_key='default'

    ###################################################
    ck.out(sep)
    ck.out('Checking available data sets ...')

    dsets=i.get('dataset_uoas',[])

    dtags=pd.get('run_cmds',{}).get(cmd_key,{}).get('dataset_tags','')

    ii={'action':'search',
        'module_uoa':cfg['module_deps']['dataset']}
    if len(dsets)>0:
       ii['data_uoa_list']=dsets
    else:
       ii['tags']=dtags
    r=ck.access(ii)
    if r['return']>0: return r
    dlist=r['lst']

    # Prepare first and second line of table
    t=[]
    t.append('')
    t.append('')
    t.append('')
    for ds in dlist:
        t.append('Dataset '+ds['data_uoa']+':')
    table.append(t)

    t=[]
    t.append('Optimization:')
    t.append('Binary size:')
    t.append('MD5SUM:')
    for ds in dlist:
        t.append('min time (s); exp time (s); var (%):')
    table.append(t)

    # Number of statistical repetitions
    srepeat=int(i.get('stat_repeat',0))
    if srepeat<1: srepeat=4

    repeat=i.get('repeat',-1)

    hos=i.get('host_os','')
    tos=i.get('target_os','')
    tdid=i.get('device_id','')

    # will be updated later
    deps={}
    features={}
    xchoices={}

    dcomp=''

    for cf in cflags:
        ck.out(sep)
        ck.out('Checking flags "'+cf+'" ...')

        t=[]
        t.append(cf)

        ii={'action':'run',

            'module_uoa':cfg['module_deps']['pipeline'],
            'data_uoa':cfg['module_deps']['program'],

            'host_os':hos,
            'target_os':tos,
            'device_id':tdid,

            'repetitions': 1,

            'force_pipeline_update':'yes',

            'out':'con'}

        pipeline={'data_uoa':puoa,
                  'flags':cf,
                  'no_run':'yes'}

        if len(ap)>0: pipeline.update(ap)

        if len(deps)>0: pipeline['dependencies']=deps

        ii['pipeline']=pipeline

        r=ck.access(ii)
        if r['return']>0: return r

        lio=r.get('last_iteration_output',{})

        fail=lio.get('fail','')

        if fail=='yes':
           return {'return':1, 'error':'compilation failed ('+lio.get('fail_reason','')+')- check above output and possibly report to the authors!'}

        ed=r.get('experiment_desc',{})
        deps=ed.get('dependencies',{})

        cc=ed.get('choices',{})

        if len(xchoices)==0: xchoices=cc

        hos=cc['host_os']
        tos=cc['target_os']
        tdid=cc['device_id']

        ft=ed.get('features',{})
        if len(features)==0: features=ft

        if dcomp=='': dcomp=ft.get('compiler_version',{}).get('str','')

        lsa=r.get('last_stat_analysis',{})
        fresults=lsa.get('dict_flat',{})

        os=fresults.get('##characteristics#compile#binary_size#min',0)
        md5=fresults.get('##characteristics#compile#md5_sum#min','')

        t.append(os)
        t.append(md5)

        # Iterate over datasets
        oresults={}

        for ds in dlist:
            duoa=ds['data_uoa']
            duid=ds['data_uid']

            ck.out(sep)
            ck.out('Running with dataset '+duoa+' ...')
            ck.out('')

            ij={'action':'run',

                'module_uoa':cfg['module_deps']['pipeline'],
                'data_uoa':cfg['module_deps']['program'],
                'program_uoa':puoa,

                'host_os':hos,
                'target_os':tos,
                'device_id':tdid,

                'repetitions': srepeat,

                'force_pipeline_update':'yes',

                'out':'con'}

            pipeline={'data_uoa':puoa,
                      'cmd_key':cmd_key,
                      'dataset_uoa':duid,
                      'no_compile':'yes'}

            if len(ap)>0: pipeline.update(ap)

            if len(deps)>0: pipeline['dependencies']=deps

            if repeat>0: pipeline['repeat']=repeat

            ij['pipeline']=pipeline

            r=ck.access(ij)
            if r['return']>0: return r

            lio=r.get('last_iteration_output',{})

            fail=lio.get('fail','')

            if fail=='yes':
               return {'return':1, 'error':'execution failed ('+lio.get('fail_reason','')+')- check above output and possibly report to the authors!'}

            state=lio.get('state',{})
            repeat=state['repeat']

            lsa=r.get('last_stat_analysis',{})
            fresults=lsa.get('dict_flat',{})

            texp=fresults.get('##characteristics#run#execution_time_kernel_0#exp',0)
            tmin=fresults.get('##characteristics#run#execution_time_kernel_0#min',0)
            tdelta=fresults.get('##characteristics#run#execution_time_kernel_0#range_percent',0)

            oresults[duoa]=fresults

            t.append('      '+('%3.3f' % tmin) + ' ;       ' + ('%3.3f' % texp) + ' ;   ' + ('%4.1f' % (tdelta*100))+'%')

        otable.append(oresults)
        table.append(t)

    # Draw table
    ii={'action':'draw',
        'module_uoa':cfg['module_deps']['table'],
        'table':table,
        'out':'txt'}
    r=ck.access(ii)
    if r['return']>0: return r
    s=r['string']

    rft=rf+'.txt'
    rfh=rf+'.html'
    rfj=rf+'.json'

    ck.out(sep)
    ck.out('Raw results (exported to '+rf+'.txt, .html, .json):')

    if dcomp!='':
       ck.out('')
       ck.out('Detected compiler version: '+dcomp)

    ck.out('')
    ck.out(s)

    r=ck.save_text_file({'text_file':rft, 'string':s})
    if r['return']>0: return r

    ii['out']='html'
    r=ck.access(ii)
    if r['return']>0: return r
    html=r['string']

    r=ck.save_text_file({'text_file':rfh, 'string':html})
    if r['return']>0: return r

    # Checking if there is a speedup ...
    # Expect that otable[0] - -O3; otable[1] - -O3 -fno-if-conversion
    if i.get('check_speedup','')=='yes' and len(otable)>1:
       r0d0=otable[0][dlist[0]['data_uoa']]
       r0d1=otable[0][dlist[1]['data_uoa']]
       r1d0=otable[1][dlist[0]['data_uoa']]
       r1d1=otable[1][dlist[1]['data_uoa']]

#       t0d0=r0d0['##characteristics#run#execution_time_kernel_0#exp']/repeat
#       t0d1=r0d1['##characteristics#run#execution_time_kernel_0#exp']/repeat
#       t1d0=r1d0['##characteristics#run#execution_time_kernel_0#exp']/repeat
#       t1d1=r1d1['##characteristics#run#execution_time_kernel_0#exp']/repeat

       t0d0=r0d0['##characteristics#run#execution_time_kernel_0#min']/repeat
       t0d1=r0d1['##characteristics#run#execution_time_kernel_0#min']/repeat
       t1d0=r1d0['##characteristics#run#execution_time_kernel_0#min']/repeat
       t1d1=r1d1['##characteristics#run#execution_time_kernel_0#min']/repeat

       sd0=t0d0/t1d0
       sd1=t0d1/t1d1

       if sd0>1.08 or sd1>1.08 or sd0<0.92 or sd1<0.92 or i.get('force_record','')=='yes':
          ck.out(sep)
          ck.out('Found speedup or slow down for the first 2 optimizations:')
          ck.out('')
          ck.out('* Dataset 0 ('+dlist[0]['data_uoa']+') speedup (T_opt0/T_opt1) = '+('%2.2f' % sd0))
          ck.out('* Dataset 1 ('+dlist[1]['data_uoa']+') speedup (T_opt0/T_opt1) = '+('%2.2f' % sd1))

          ck.out('')
          r=ck.inp({'text':'Would you like to share this result with the community and author via public "remote-ck" web service (Y/n): '})
          x=r['string'].lower()
          if x=='' or x=='yes' or x=='y':
             xchoices['optimization_0']=cflags[0]
             xchoices['optimization_1']=cflags[1]

             xchoices['dataset_uoa_0']=dlist[0]['data_uoa']
             xchoices['dataset_uoa_1']=dlist[1]['data_uoa']

             xchoices['dataset_uid_0']=dlist[0]['data_uid']
             xchoices['dataset_uid_1']=dlist[1]['data_uid']

             xchoices['compiler_version']=dcomp

             ii={'action':'add',
                 'module_uoa':cfg['module_deps']['experiment'],

                 'repo_uoa':e_repo_uoa,
                 'experiment_repo_uoa':e_remote_repo_uoa,
                 'experiment_uoa':e_uoa,

                 'sort_keys':'yes',

                 'dict':{
                   'dict':{'subview_uoa':cfg['data_deps']['subview_uoa']},

                   'tags':['crowdsource experiments','ck-paper','filter','if-conversion','speedup'],

                   'features':features,
                   'choices':xchoices,
                   'characteristics': {
                      'speedup_0':sd0,
                      'speedup_1':sd1,
                      'execution_time_div_by_repeat_opt0_ds0':t0d0,
                      'execution_time_div_by_repeat_opt0_ds1':t0d1,
                      'execution_time_div_by_repeat_opt1_ds0':t1d0,
                      'execution_time_div_by_repeat_opt1_ds1':t1d1
                   }
                 }
                }

             r=ck.access(ii)
             if r['return']>0: return r

             ck.out('')
             ck.out('  Results shared successfully!')

             ck.out('')
             ck.out('  You can see all shared results at http://cknowledge.org/repo/web.php?wcid=bc0409fb61f0aa82:reproduce-ck-paper-filter-optimization')

       else:
          ck.out('')
          ck.out('Note: speedups/slowdowns were not detected on your platform!')

    ck.out('')
    ck.out('Thank you for participating in experiment crowdsourcing!')

    return {'return':0}



#
# Collective Knowledge (experiment participants and their stats)
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# show users

def show(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0

              html         - generated HTML
            }

    """

    import os

    h='<center>'
    h+='<h2>Collective Knowledge Timeline<BR>(non-anonymous participation in collaborative experiments)</h2>\n'

    uu=[]

    # Load all
    rx=ck.access({'action':'load',
                  'module_uoa':work['self_module_uid'],
                  'data_uoa':'all'})
    if rx['return']>0:
       if rx['return']!=16:
          return rx
       else:
          h+='<b>Timeline is empty ...</b>'
    else:
       d=rx['dict']

       du=d.get('users',{})
       dt=d.get('timeline',[])

       for q in dt:
           user=q.get('user','')
           if user!='' and user not in uu:
              uu.append(user)

       unique_users=len(uu)

       h+='<i>Number of unique users: '+str(unique_users)+'</i><br><br>\n'

       # Check host URL prefix and default module/action
       rx=ck.access({'action':'form_url_prefix',
                     'module_uoa':'wfe',
                     'host':i.get('host',''), 
                     'port':i.get('port',''), 
                     'template':i.get('template','')})
       if rx['return']>0: return rx
       url0=rx['url']

       xdt=sorted(dt, key=lambda v: (v.get('iso_datetime','')))

       h+='<center>\n'
       h+='<table class="ck_table" border="0" cellpadding="5">\n'

       h+=' <tr style="background-color:#cfcfff;">\n'
       h+='  <td><b>\n'
       h+='   Date / time\n'
       h+='  </b></td>\n'

       h+='  <td><b>\n'
       h+='   User\n'
       h+='  </b></td>\n'

       h+='  <td align="center"><b>\n'
       h+='   Action\n'
       h+='  </b></td>\n'

       h+='  <td align="center"><b>\n'
       h+='   Scenario\n'
       h+='  </b></td>\n'

       h+=' </tr>\n'

       for q in dt:
           idt=q.get('iso_datetime','')
           nu=q.get('new_user','')
           user=q.get('user','')
           tp=q.get('type','')

           ix=user.find('@')
           if ix>1:
              user=user[:ix-1]+'{HiddenForPrivacy}'+user[ix+2:]

           a=''
           if nu=='yes': a='New user'
           else:
              suid=q.get('solution_uid','')
              scenario_uoa=q.get('scenario_uoa','')
              duoa=q.get('data_uoa','')

              a='<a href="'+url0+'wcid='+scenario_uoa+':'+duoa+'">Better solution found (UID='+suid+')</a>'

           h+='<tr>'

           h+=' <td align="left">'+idt.replace('T',' ')+'</td>'
           h+=' <td align="left">'+user+'</td>'
           h+=' <td align="left">'+a+'</td>'
           h+=' <td align="left">'+tp+'</td>'

           h+='</tr>'

       h+='</table>\n'

       h+='</center>\n'

    return {'return':0, 'html':h}



#
# Collective Knowledge ()
#
# See CK LICENSE.txt for licensing details
# See CK COPYRIGHT.txt for copyright details
#
# Developer: Grigori Fursin, Grigori.Fursin@cTuning.org, http://fursin.net
#

cfg={}  # Will be updated by CK (meta description of this module)
work={} # Will be updated by CK (temporal data)
ck=None # Will be updated by CK (initialized CK kernel) 

# Local settings

##############################################################################
# Initialize module

def init(i):
    """

    Input:  {}

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """
    return {'return':0}

##############################################################################
# joke

def make(i):
    """
    Input:  {
            }

    Output: {
              return       - return code =  0, if successful
                                         >  0, if error
              (error)      - error text if return > 0
            }

    """

    cids=i.get('cids',[])

    x=''
    if len(cids)>0:
       x=cids[0]

    if x=='':
       ck.out('happy ...')

    elif x=='happy':
       ck.out('Of course!')

    elif x=='unhappy':
       ck.out('What??????')

    else:
       ck.out('Are you sure? Ok, CK will try to make you '+x+' ...')

    ck.out('')
    ck.out('CK rules:')
    ck.out('')
    ck.out('1) CK always makes you happy!')
    ck.out('2) If CK doesn\'t make you happy - see rule 1')

    return {'return':0}
